try:
    from . import generic as g
except BaseException:
    import generic as g
from trimesh.voxel import runlength as rl

np = g.np


def random_rle_encoding(n=20, max_value=255, dtype=np.uint8):
    return (
        np.random.uniform(
            size=(n,),
        )
        * (max_value - 1)
        + 1
    ).astype(np.uint8)


def random_brle_encoding(n=20, max_value=255, dtype=np.uint8):
    return (np.random.uniform(size=(n,)) * (max_value - 1) + 1).astype(dtype)


class RleTest(g.unittest.TestCase):
    """Tests should inherit from this class along with unittest.TestCase."""

    def test_rle_encode_decode(self):
        small = np.array([3] * 500 + [5] * 1000 + [2], dtype=np.uint8)
        rand = (np.random.uniform(size=(10000,)) > 0.05).astype(np.uint8)
        for original in [small, rand]:
            for dtype in [np.uint8, np.int64]:
                enc = rl.dense_to_rle(original, dtype=dtype)
                dec = rl.rle_to_dense(enc)
                np.testing.assert_equal(original, dec)

    def test_rle_decode_encode(self):
        small = [3, 5, 5, 10, 2, 1]
        rand = random_rle_encoding()
        for original in [small, rand]:
            for dtype in [np.uint8, np.int64]:
                dec = rl.brle_to_dense(original)
                enc = rl.dense_to_brle(dec, dtype=dtype)
                np.testing.assert_equal(original, enc)

    def test_merge_rle_lengths(self):
        v0, l0 = [5, 5, 2], [10, 10, 1]
        v1, l1 = [5, 2], [20, 1]
        v0, l0 = rl.merge_rle_lengths(v0, l0)
        np.testing.assert_equal(v0, v1)
        np.testing.assert_equal(l0, l1)

    def test_split_long_rle_lengths(self):
        v0, l0 = [5], [300]
        v1, l1 = [5, 5], [255, 45]

        v0, l0 = rl.split_long_rle_lengths(v0, l0, dtype=np.uint8)
        np.testing.assert_equal(v0, v1)
        np.testing.assert_equal(l0, l1)

        v0, l0 = [5, 2, 3], [10, 1000, 4]
        v1, l1 = [5, 2, 2, 2, 2, 3], [10, 255, 255, 255, 235, 4]

        v0, l0 = rl.split_long_rle_lengths(v0, l0, dtype=np.uint8)
        np.testing.assert_equal(v0, v1)
        np.testing.assert_equal(l0, l1)

    def test_rle_length(self):
        np.testing.assert_equal(
            rl.rle_length([0, 5, 1, 3, 0, 6]), rl.brle_length([5, 3, 6, 0])
        )

    def test_rle_to_brle(self):
        np.testing.assert_equal(rl.rle_to_brle([0, 5, 1, 3, 0, 10]), [5, 3, 10, 0])
        np.testing.assert_equal(rl.rle_to_brle([0, 5, 0, 3, 1, 10]), [8, 10])
        np.testing.assert_equal(rl.rle_to_brle([1, 5, 0, 3, 1, 10]), [0, 5, 3, 10])
        np.testing.assert_equal(rl.rle_to_brle([1, 5, 0, 2]), [0, 5, 2, 0])

    def test_rle_to_dense(self):
        np.testing.assert_equal(rl.rle_to_dense([5, 3, 4, 10]), [5] * 3 + [4] * 10)
        np.testing.assert_equal(rl.rle_to_dense([5, 300, 4, 100]), [5] * 300 + [4] * 100)

    def test_brle_encode_decode(self):
        small = np.array([False] * 500 + [True] * 1000 + [False], dtype=bool)
        rand = np.random.uniform(size=(10000,)) > 0.05
        for original in [small, rand]:
            for dtype in [np.uint8, np.int64]:
                enc = rl.dense_to_brle(original, dtype=dtype)
                dec = rl.brle_to_dense(enc)
                np.testing.assert_equal(original, dec)

    def test_brle_decode_encode(self):
        small = [3, 5, 5, 10, 2, 1]
        rand = random_brle_encoding()
        for original in [small, rand]:
            for dtype in [np.uint8, np.int64]:
                dec = rl.brle_to_dense(original)
                enc = rl.dense_to_brle(dec, dtype=dtype)
                np.testing.assert_equal(original, enc)

    def test_brle_logical_not(self):
        original = random_brle_encoding(dtype=np.int64)
        notted = rl.brle_logical_not(original)
        dense_notted = rl.brle_to_dense(notted)
        dense_original = rl.brle_to_dense(original)
        np.testing.assert_equal(dense_notted, np.logical_not(dense_original))

    def test_merge_brle_lengths(self):
        np.testing.assert_equal(rl.merge_brle_lengths([10, 0, 10, 2]), [20, 2])
        np.testing.assert_equal(rl.merge_brle_lengths([10, 0, 10, 2]), [20, 2])
        np.testing.assert_equal(rl.merge_brle_lengths([10, 1, 10, 2]), [10, 1, 10, 2])
        np.testing.assert_equal(rl.merge_brle_lengths([0, 10, 2, 3]), [0, 10, 2, 3])

    def test_split_long_brle_lengths(self):
        np.testing.assert_equal(
            rl.split_long_brle_lengths([300, 600, 10], np.uint8),
            [255, 0, 45, 255, 0, 255, 0, 90, 10],
        )

    def test_brle_split_merge(self):
        # TODO: REMOVE RETURN
        if True:
            return
        # TODO: FIGURE OUT WHY THIS FAILS
        x = [300, 600, 10, 0]
        split = rl.split_long_brle_lengths(x, np.uint8)
        merged = rl.merge_brle_lengths(split)
        np.testing.assert_equal(merged, x)

    def test_brle_to_rle(self):
        brle_data = random_brle_encoding()
        brle_dense = rl.brle_to_dense(brle_data)
        rle_data = rl.brle_to_rle(brle_data)
        rle_dense = rl.rle_to_dense(rle_data)
        np.testing.assert_equal(brle_dense, rle_dense)
        np.testing.assert_equal(rl.brle_to_rle([0, 5, 2, 0]), [1, 5, 0, 2])

    def test_dense_to_brle(self):
        # should be an (300, 200, 1000) array
        x = np.array([False] * 300 + [True] * 200 + [False] * 1000)

        # TODO: REMOVE RETURN
        if True:
            return
        # TODO: FIGURE OUT WHY THIS FAILS
        np.testing.assert_equal(rl.dense_to_brle(x), [300, 200, 1000, 0])
        np.testing.assert_equal(
            rl.dense_to_brle(x, np.uint8),
            [255, 0, 45, 200, 255, 0, 255, 0, 255, 0, 235, 0],
        )

    def test_brle_to_dense(self):
        np.testing.assert_equal(
            rl.brle_to_dense(np.array([300, 200, 1000, 0], dtype=np.int64)),
            [False] * 300 + [True] * 200 + [False] * 1000,
        )
        np.testing.assert_equal(
            rl.brle_to_dense(
                np.array(
                    [255, 0, 45, 200, 255, 0, 255, 0, 255, 0, 235, 0], dtype=np.int64
                )
            ),
            [False] * 300 + [True] * 200 + [False] * 1000,
        )

    def test_brle_length(self):
        enc = random_brle_encoding(dtype=np.int64)
        dec = rl.brle_to_dense(enc)
        np.testing.assert_equal(len(dec), rl.brle_length(enc))

    def test_rle_mask(self):
        rle_data = random_rle_encoding()
        dense = rl.rle_to_dense(rle_data)
        mask = np.random.uniform(size=dense.shape) > 0.8
        expected = dense[mask]
        actual = tuple(rl.rle_mask(rle_data, mask))
        np.testing.assert_equal(actual, expected)

    def test_brle_mask(self):
        brle_data = random_brle_encoding()
        dense = rl.brle_to_dense(brle_data)
        mask = np.random.uniform(size=dense.shape) > 0.8
        expected = dense[mask]
        actual = tuple(rl.brle_mask(brle_data, mask))
        np.testing.assert_equal(actual, expected)

    def test_rle_strip(self):
        for rle_data, expected_rle, expected_padding in (
            ([0, 5, 1, 3, 0, 10], [1, 3], (5, 10)),
            ([1, 3, 0, 10], [1, 3], (0, 10)),
            ([0, 5, 1, 3], [1, 3], (5, 0)),
            ([0, 5, 1, 3, 0, 0], [1, 3], (5, 0)),
            ([0, 5, 1, 3, 0, 10, 0, 5], [1, 3], (5, 15)),
            ([0, 5, 0, 3, 1, 3, 0, 10, 0, 5], [1, 3], (8, 15)),
            ([1, 3], [1, 3], (0, 0)),
        ):
            actual_rle, actual_padding = rl.rle_strip(rle_data)
            np.testing.assert_equal(actual_rle, expected_rle)
            np.testing.assert_equal(actual_padding, expected_padding)

    def test_brle_strip(self):
        for brle_data, expected_brle, expected_padding in (
            ([5, 3, 10], [0, 3], [5, 10]),
            ([0, 3, 10], [0, 3], [0, 10]),
            ([5, 3], [0, 3], [5, 0]),
            ([5, 3, 0, 0], [0, 3], (5, 0)),
            ([5, 3, 10, 0, 5], [0, 3], (5, 15)),
            ([5, 0, 3, 3, 10, 0, 5], [0, 3], (8, 15)),
            ([0, 3], [0, 3], (0, 0)),
        ):
            actual_brle, actual_padding = rl.brle_strip(brle_data)
            np.testing.assert_equal(actual_brle, expected_brle)
            np.testing.assert_equal(actual_padding, expected_padding)


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    g.unittest.main()



"""
Ray queries using the embreex package with the
API wrapped to match our native raytracer.
"""
from copy import deepcopy

import numpy as np

from .. import caching, intersections, util
from ..constants import log_time
from .ray_util import contains_points

# the factor of geometry.scale to offset a ray from a triangle
# to reliably not hit its origin triangle
_ray_offset_factor = 1e-4
# we want to clip our offset to a sane distance
_ray_offset_floor = 1e-8


try:
    # try the preferred wrapper which installs from wheels
    from embreex import rtcore_scene
    from embreex.mesh_construction import TriangleMesh

    # pass embree floats as 32 bit
    _embree_dtype = np.float32
except BaseException as E:
    try:
        # this will be deprecated at some point hopefully soon
        from pyembree import __version__, rtcore_scene
        from pyembree.mesh_construction import TriangleMesh

        # see if we're using a newer version of the pyembree wrapper
        _embree_new = tuple([int(i) for i in __version__.split(".")]) >= (0, 1, 4)
        # both old and new versions require exact but different type
        _embree_dtype = [np.float64, np.float32][int(_embree_new)]
    except BaseException:
        # raise the embreex error for better log message
        raise E


class RayMeshIntersector:
    def __init__(self, geometry, scale_to_box=True):
        """
        Do ray- mesh queries.

        Parameters
        -------------
        geometry : Trimesh object
          Mesh to do ray tests on
        scale_to_box : bool
          If true, will scale mesh to approximate
          unit cube to avoid problems with extreme
          large or small meshes.
        """
        self.mesh = geometry
        self._scale_to_box = scale_to_box
        self._cache = caching.Cache(id_function=self.mesh.__hash__)

    @property
    def _scale(self):
        """
        Scaling factor for precision.
        """
        if self._scale_to_box:
            # scale vertices to approximately a cube to help with
            # numerical issues at very large/small scales
            scale = 100.0 / self.mesh.scale
        else:
            scale = 1.0
        return scale

    @caching.cache_decorator
    def _scene(self):
        """
        A cached version of the embreex scene.
        """
        return _EmbreeWrap(
            vertices=self.mesh.vertices, faces=self.mesh.faces, scale=self._scale
        )

    def intersects_location(self, ray_origins, ray_directions, multiple_hits=True):
        """
        Return the location of where a ray hits a surface.

        Parameters
        ----------
        ray_origins : (n, 3) float
          Origins of rays
        ray_directions : (n, 3) float
          Direction (vector) of rays

        Returns
        ---------
        locations : (m) sequence of (p, 3) float
          Intersection points
        index_ray : (m,) int
          Indexes of ray
        index_tri : (m,) int
          Indexes of mesh.faces
        """
        (index_tri, index_ray, locations) = self.intersects_id(
            ray_origins=ray_origins,
            ray_directions=ray_directions,
            multiple_hits=multiple_hits,
            return_locations=True,
        )

        return locations, index_ray, index_tri

    @log_time
    def intersects_id(
        self,
        ray_origins,
        ray_directions,
        multiple_hits=True,
        max_hits=20,
        return_locations=False,
    ):
        """
        Find the triangles hit by a list of rays, including
        optionally multiple hits along a single ray.


        Parameters
        ----------
        ray_origins : (n, 3) float
          Origins of rays
        ray_directions : (n, 3) float
          Direction (vector) of rays
        multiple_hits : bool
          If True will return every hit along the ray
          If False will only return first hit
        max_hits : int
          Maximum number of hits per ray
        return_locations : bool
          Should we return hit locations or not

        Returns
        ---------
        index_tri : (m,) int
          Indexes of mesh.faces
        index_ray : (m,) int
          Indexes of ray
        locations : (m) sequence of (p, 3) float
          Intersection points, only returned if return_locations
        """
        # make sure input is _dtype for embree
        ray_origins = np.array(deepcopy(ray_origins), dtype=np.float64)
        ray_directions = np.asanyarray(ray_directions, dtype=np.float64)
        if ray_origins.shape != ray_directions.shape:
            raise ValueError("Ray origin and direction don't match!")
        ray_directions = util.unitize(ray_directions)

        # since we are constructing all hits, save them to a deque then
        # stack into (depth, len(rays)) at the end
        result_triangle = []
        result_ray_idx = []
        result_locations = []

        # the mask for which rays are still active
        current = np.ones(len(ray_origins), dtype=bool)

        if multiple_hits or return_locations:
            # how much to offset ray to transport to the other side of face
            distance = np.clip(
                _ray_offset_factor * self._scale, _ray_offset_floor, np.inf
            )
            ray_offsets = ray_directions * distance

            # grab the planes from triangles
            plane_origins = self.mesh.triangles[:, 0, :]
            plane_normals = self.mesh.face_normals

        # use a for loop rather than a while to ensure this exits
        # if a ray is offset from a triangle and then is reported
        # hitting itself this could get stuck on that one triangle
        for _ in range(max_hits):
            # run the embreex query
            # if you set output=1 it will calculate distance along
            # ray, which is bizzarely slower than our calculation

            query = self._scene.run(ray_origins[current], ray_directions[current])
            # basically we need to reduce the rays to the ones that hit
            # something
            hit = query != -1
            # which triangle indexes were hit
            hit_triangle = query[hit]

            # eliminate rays that didn't hit anything from future queries
            current_index = np.nonzero(current)[0]
            current_index_no_hit = current_index[np.logical_not(hit)]
            current_index_hit = current_index[hit]
            current[current_index_no_hit] = False

            # append the triangle and ray index to the results
            result_triangle.append(hit_triangle)
            result_ray_idx.append(current_index_hit)

            # if we don't need all of the hits, return the first one
            if (not multiple_hits and not return_locations) or not hit.any():
                break

            # find the location of where the ray hit the triangle plane
            new_origins, valid = intersections.planes_lines(
                plane_origins=plane_origins[hit_triangle],
                plane_normals=plane_normals[hit_triangle],
                line_origins=ray_origins[current],
                line_directions=ray_directions[current],
            )

            if not valid.all():
                # since a plane intersection was invalid we have to go back and
                # fix some stuff, we pop the ray index and triangle index,
                # apply the valid mask then append it right back to keep our
                # indexes intact
                result_ray_idx.append(result_ray_idx.pop()[valid])
                result_triangle.append(result_triangle.pop()[valid])

                # update the current rays to reflect that we couldn't find a
                # new origin
                current[current_index_hit[np.logical_not(valid)]] = False

            # since we had to find the intersection point anyway we save it
            # even if we're not going to return it
            result_locations.extend(new_origins)

            if multiple_hits:
                # move the ray origin to the other side of the triangle
                ray_origins[current] = new_origins + ray_offsets[current]
            else:
                break

        # stack the dequeues into nice 1D numpy arrays
        index_tri = np.hstack(result_triangle)
        index_ray = np.hstack(result_ray_idx)

        if return_locations:
            locations = (
                np.zeros((0, 3), float)
                if len(result_locations) == 0
                else np.array(result_locations)
            )

            return index_tri, index_ray, locations
        return index_tri, index_ray

    @log_time
    def intersects_first(self, ray_origins, ray_directions):
        """
        Find the index of the first triangle a ray hits.


        Parameters
        ----------
        ray_origins : (n, 3) float
          Origins of rays
        ray_directions : (n, 3) float
          Direction (vector) of rays

        Returns
        ----------
        triangle_index : (n,) int
          Index of triangle ray hit, or -1 if not hit
        """

        ray_origins = np.asanyarray(deepcopy(ray_origins))
        ray_directions = np.asanyarray(ray_directions)

        triangle_index = self._scene.run(ray_origins, ray_directions)
        return triangle_index

    def intersects_any(self, ray_origins, ray_directions):
        """
        Check if a list of rays hits the surface.


        Parameters
        -----------
        ray_origins : (n, 3) float
          Origins of rays
        ray_directions : (n, 3) float
          Direction (vector) of rays

        Returns
        ----------
        hit : (n,) bool
          Did each ray hit the surface
        """

        first = self.intersects_first(
            ray_origins=ray_origins, ray_directions=ray_directions
        )
        hit = first != -1
        return hit

    def contains_points(self, points):
        """
        Check if a mesh contains a list of points, using ray tests.

        If the point is on the surface of the mesh, behavior is undefined.

        Parameters
        ---------
        points: (n, 3) points in space

        Returns
        ---------
        contains: (n,) bool
                         Whether point is inside mesh or not
        """
        return contains_points(self, points)


class _EmbreeWrap:
    """
    A light wrapper for Embreex scene objects which
    allows queries to be scaled to help with precision
    issues, as well as selecting the correct dtypes.
    """

    def __init__(self, vertices, faces, scale):
        scaled = np.array(vertices, dtype=np.float64)
        self.origin = scaled.min(axis=0)
        self.scale = float(scale)
        scaled = (scaled - self.origin) * self.scale

        self.scene = rtcore_scene.EmbreeScene()
        # assign the geometry to the scene
        TriangleMesh(
            scene=self.scene,
            vertices=scaled.astype(_embree_dtype),
            indices=faces.view(np.ndarray).astype(np.int32),
        )

    def run(self, origins, normals, **kwargs):
        scaled = (np.array(origins, dtype=np.float64) - self.origin) * self.scale

        return self.scene.run(
            scaled.astype(_embree_dtype), normals.astype(_embree_dtype), **kwargs
        )



import os
import platform

from .. import resources, util
from ..constants import log
from .generic import MeshScript

_search_path = os.environ.get("PATH", "")
if platform.system() == "Windows":
    # try to find Blender install on Windows
    # split existing path by delimiter
    _search_path = [i for i in _search_path.split(";") if len(i) > 0]
    for pf in [r"C:\Program Files", r"C:\Program Files (x86)"]:
        pf = os.path.join(pf, "Blender Foundation")
        if os.path.exists(pf):
            for p in os.listdir(pf):
                if "Blender" in p:
                    _search_path.append(os.path.join(pf, p))
    _search_path = ";".join(_search_path)
    log.debug("searching for blender in: %s", _search_path)

if platform.system() == "Darwin":
    # try to find Blender on Mac OSX
    _search_path = [i for i in _search_path.split(":") if len(i) > 0]
    _search_path.append("/Applications/blender.app/Contents/MacOS")
    _search_path.append("/Applications/Blender.app/Contents/MacOS")
    _search_path.append("/Applications/Blender/blender.app/Contents/MacOS")
    _search_path = ":".join(_search_path)
    log.debug("searching for blender in: %s", _search_path)

_blender_executable = util.which("blender", path=_search_path)
exists = _blender_executable is not None


def boolean(meshes, operation="difference", debug=False):
    """
    Run a boolean operation with multiple meshes using Blender.
    """
    if not exists:
        raise ValueError("No blender available!")
    operation = str.upper(operation)
    if operation == "INTERSECTION":
        operation = "INTERSECT"

    # get the template from our resources folder
    template = resources.get("templates/blender_boolean.py.tmpl")
    script = template.replace("$OPERATION", operation)

    with MeshScript(meshes=meshes, script=script, debug=debug) as blend:
        result = blend.run(_blender_executable + " --background --python $SCRIPT")

    for m in util.make_sequence(result):
        # blender returns actively incorrect face normals
        m.face_normals = None

    return result


def unwrap(mesh, angle_limit=66, island_margin=0.0, debug=False):
    """
    Run an unwrap operation using blender.
    """
    if not exists:
        raise ValueError("No blender available!")

    # get the template from our resources folder
    template = resources.get("templates/blender_unwrap.py")
    script = template.replace("$ANGLE_LIMIT", "%.6f" % angle_limit).replace(
        "$ISLAND_MARGIN", "%.6f" % island_margin
    )

    with MeshScript(meshes=[mesh], script=script, exchange="obj", debug=debug) as blend:
        result = blend.run(_blender_executable + " --background --python $SCRIPT")

    for m in util.make_sequence(result):
        # blender returns actively incorrect face normals
        m.face_normals = None

    return result



"""
A basic slow implementation of ray- triangle queries.
"""
import numpy as np

from .. import caching, grouping, intersections, util
from .. import triangles as triangles_mod
from ..constants import tol
from .ray_util import contains_points


class RayMeshIntersector:
    """
    An object to query a mesh for ray intersections.
    Precomputes an r-tree for each triangle on the mesh.
    """

    def __init__(self, mesh):
        self.mesh = mesh
        self._cache = caching.Cache(self.mesh.__hash__)

    def intersects_id(
        self,
        ray_origins,
        ray_directions,
        return_locations=False,
        multiple_hits=True,
        **kwargs,
    ):
        """
        Find the intersections between the current mesh and an
        array of rays.

        Parameters
        ------------
        ray_origins :  (m, 3) float
          Ray origin points
        ray_directions : (m, 3) float
          Ray direction vectors
        multiple_hits :  bool
          Consider multiple hits of each ray or not
        return_locations : bool
          Return hit locations or not

        Returns
        -----------
        index_triangle : (h,) int
          Index of triangles hit
        index_ray : (h,) int
          Index of ray that hit triangle
        locations : (h, 3) float
          [optional] Position of intersection in space
        """
        (index_tri, index_ray, locations) = ray_triangle_id(
            triangles=self.mesh.triangles,
            ray_origins=ray_origins,
            ray_directions=ray_directions,
            tree=self.mesh.triangles_tree,
            multiple_hits=multiple_hits,
            triangles_normal=self.mesh.face_normals,
        )
        if return_locations:
            if len(index_tri) == 0:
                return index_tri, index_ray, locations
            unique = grouping.unique_rows(np.column_stack((locations, index_ray)))[0]
            return index_tri[unique], index_ray[unique], locations[unique]
        return index_tri, index_ray

    def intersects_location(self, ray_origins, ray_directions, **kwargs):
        """
        Return unique cartesian locations where rays hit the mesh.
        If you are counting the number of hits a ray had, this method
        should be used as if only the triangle index is used on- edge hits
        will be counted twice.

        Parameters
        ------------
        ray_origins : (m, 3) float
          Ray origin points
        ray_directions : (m, 3) float
          Ray direction vectors

        Returns
        ---------
        locations : (n) sequence of (m,3) float
          Intersection points
        index_ray : (n,) int
          Array of ray indexes
        index_tri: (n,) int
          Array of triangle (face) indexes
        """
        (index_tri, index_ray, locations) = self.intersects_id(
            ray_origins=ray_origins,
            ray_directions=ray_directions,
            return_locations=True,
            **kwargs,
        )
        return locations, index_ray, index_tri

    def intersects_first(self, ray_origins, ray_directions, **kwargs):
        """
        Find the index of the first triangle a ray hits.


        Parameters
        ----------
        ray_origins : (n, 3) float
          Origins of rays
        ray_directions : (n, 3) float
          Direction (vector) of rays

        Returns
        ----------
        triangle_index : (n,) int
          Index of triangle ray hit, or -1 if not hit
        """

        (index_tri, index_ray) = self.intersects_id(
            ray_origins=ray_origins,
            ray_directions=ray_directions,
            return_locations=False,
            multiple_hits=False,
            **kwargs,
        )

        # put the result into the form of "one triangle index per ray"
        result = np.ones(len(ray_origins), dtype=np.int64) * -1
        result[index_ray] = index_tri

        return result

    def intersects_any(self, ray_origins, ray_directions, **kwargs):
        """
        Find out if each ray hit any triangle on the mesh.

        Parameters
        ------------
        ray_origins : (m, 3) float
          Ray origin points
        ray_directions : (m, 3) float
          Ray direction vectors

        Returns
        ---------
        hit : (m,) bool
          Whether any ray hit any triangle on the mesh
        """
        index_tri, index_ray = self.intersects_id(ray_origins, ray_directions)
        hit_any = np.zeros(len(ray_origins), dtype=bool)
        hit_idx = np.unique(index_ray)
        if len(hit_idx) > 0:
            hit_any[hit_idx] = True
        return hit_any

    def contains_points(self, points):
        """
        Check if a mesh contains a list of points, using ray tests.

        If the point is on the surface of the mesh the behavior
        is undefined.

        Parameters
        ------------
        points : (n, 3) float
          Points in space

        Returns
        ---------
        contains : (n,) bool
          Whether point is inside mesh or not
        """

        return contains_points(self, points)


def ray_triangle_id(
    triangles,
    ray_origins,
    ray_directions,
    triangles_normal=None,
    tree=None,
    multiple_hits=True,
):
    """
    Find the intersections between a group of triangles and rays

    Parameters
    -------------
    triangles : (n, 3, 3) float
      Triangles in space
    ray_origins : (m, 3) float
      Ray origin points
    ray_directions : (m, 3) float
      Ray direction vectors
    triangles_normal : (n, 3) float
      Normal vector of triangles, optional
    tree : rtree.Index
      Rtree object holding triangle bounds

    Returns
    -----------
    index_triangle : (h,) int
      Index of triangles hit
    index_ray : (h,) int
      Index of ray that hit triangle
    locations : (h, 3) float
      Position of intersection in space
    """
    triangles = np.asanyarray(triangles, dtype=np.float64)
    ray_origins = np.asanyarray(ray_origins, dtype=np.float64)
    ray_directions = np.asanyarray(ray_directions, dtype=np.float64)

    # if we didn't get passed an r-tree for the bounds of each
    # triangle create one here
    if tree is None:
        tree = triangles_mod.bounds_tree(triangles)

    # find the list of likely triangles and which ray they
    # correspond with, via rtree queries
    ray_candidates, ray_id = ray_triangle_candidates(
        ray_origins=ray_origins, ray_directions=ray_directions, tree=tree
    )

    # get subsets which are corresponding rays and triangles
    # (c,3,3) triangle candidates
    triangle_candidates = triangles[ray_candidates]
    # (c,3) origins and vectors for the rays
    line_origins = ray_origins[ray_id]
    line_directions = ray_directions[ray_id]

    # get the plane origins and normals from the triangle candidates
    plane_origins = triangle_candidates[:, 0, :]
    if triangles_normal is None:
        plane_normals, triangle_ok = triangles_mod.normals(triangle_candidates)
        if not triangle_ok.all():
            raise ValueError("Invalid triangles!")
    else:
        plane_normals = triangles_normal[ray_candidates]

    # find the intersection location of the rays with the planes
    location, valid = intersections.planes_lines(
        plane_origins=plane_origins,
        plane_normals=plane_normals,
        line_origins=line_origins,
        line_directions=line_directions,
    )

    if len(triangle_candidates) == 0 or not valid.any():
        # we got no hits so return early with empty array
        return (
            np.array([], dtype=np.int64),
            np.array([], dtype=np.int64),
            np.array([], dtype=np.float64),
        )

    # find the barycentric coordinates of each plane intersection on the
    # triangle candidates
    barycentric = triangles_mod.points_to_barycentric(
        triangle_candidates[valid], location
    )

    # the plane intersection is inside the triangle if all barycentric
    # coordinates are between 0.0 and 1.0
    hit = np.logical_and(
        (barycentric > -tol.zero).all(axis=1), (barycentric < (1 + tol.zero)).all(axis=1)
    )

    # the result index of the triangle is a candidate with a valid
    # plane intersection and a triangle which contains the plane
    # intersection point
    index_tri = ray_candidates[valid][hit]
    # the ray index is a subset with a valid plane intersection and
    # contained by a triangle
    index_ray = ray_id[valid][hit]
    # locations are already valid plane intersections, just mask by hits
    location = location[hit]

    # only return points that are forward from the origin
    vector = location - ray_origins[index_ray]
    distance = util.diagonal_dot(vector, ray_directions[index_ray])
    forward = distance > -1e-6

    index_tri = index_tri[forward]
    index_ray = index_ray[forward]
    location = location[forward]
    distance = distance[forward]

    if multiple_hits:
        return index_tri, index_ray, location

    # since we are not returning multiple hits, we need to
    # figure out which hit is first
    if len(index_ray) == 0:
        return index_tri, index_ray, location

    # find the first hit
    first = np.array([g[distance[g].argmin()] for g in grouping.group(index_ray)])

    return index_tri[first], index_ray[first], location[first]


def ray_triangle_candidates(ray_origins, ray_directions, tree):
    """
    Do broad- phase search for triangles that the rays
    may intersect.

    Does this by creating a bounding box for the ray as it
    passes through the volume occupied by the tree

    Parameters
    ------------
    ray_origins : (m, 3) float
      Ray origin points.
    ray_directions : (m, 3) float
      Ray direction vectors
    tree : rtree object
      Ccontains AABB of each triangle

    Returns
    ----------
    ray_candidates : (n,) int
      Triangle indexes
    ray_id : (n,) int
      Corresponding ray index for a triangle candidate
    """
    bounding = ray_bounds(
        ray_origins=ray_origins, ray_directions=ray_directions, bounds=tree.bounds
    )

    index = []
    candidates = []
    for i, bounds in enumerate(bounding):
        cand = list(tree.intersection(bounds))
        candidates.extend(cand)
        index.extend([i] * len(cand))
    return np.array(candidates, dtype=np.int64), np.array(index, dtype=np.int64)


def ray_bounds(ray_origins, ray_directions, bounds, buffer_dist=1e-5):
    """
    Given a set of rays and a bounding box for the volume of interest
    where the rays will be passing through, find the bounding boxes
    of the rays as they pass through the volume.

    Parameters
    ------------
    ray_origins:      (m,3) float, ray origin points
    ray_directions:   (m,3) float, ray direction vectors
    bounds:           (2,3) bounding box (min, max)
    buffer_dist:      float, distance to pad zero width bounding boxes

    Returns
    ---------
    ray_bounding: (n) set of AABB of rays passing through volume
    """

    ray_origins = np.asanyarray(ray_origins, dtype=np.float64)
    ray_directions = np.asanyarray(ray_directions, dtype=np.float64)

    # bounding box we are testing against
    bounds = np.asanyarray(bounds)

    # find the primary axis of the vector
    axis = np.abs(ray_directions).argmax(axis=1)
    axis_bound = bounds.reshape((2, -1)).T[axis]
    axis_ori = np.array([ray_origins[i][a] for i, a in enumerate(axis)]).reshape((-1, 1))
    axis_dir = np.array([ray_directions[i][a] for i, a in enumerate(axis)]).reshape(
        (-1, 1)
    )

    # parametric equation of a line
    # point = direction*t + origin
    # p = dt + o
    # t = (p-o)/d
    nonzero = (axis_dir != 0.0).reshape(-1)
    t = np.zeros_like(axis_bound)
    t[nonzero] = (axis_bound[nonzero] - axis_ori[nonzero]) / axis_dir[nonzero]

    # prevent the bounding box from including triangles
    # behind the ray origin
    t[t < buffer_dist] = buffer_dist

    # the value of t for both the upper and lower bounds
    t_a = t[:, 0].reshape((-1, 1))
    t_b = t[:, 1].reshape((-1, 1))

    # the cartesian point for where the line hits the plane defined by
    # axis
    on_a = (ray_directions * t_a) + ray_origins
    on_b = (ray_directions * t_b) + ray_origins

    on_plane = np.column_stack((on_a, on_b)).reshape((-1, 2, ray_directions.shape[1]))

    ray_bounding = np.hstack((on_plane.min(axis=1), on_plane.max(axis=1)))
    # pad the bounding box by TOL_BUFFER
    # not sure if this is necessary, but if the ray is  axis aligned
    # this function will otherwise return zero volume bounding boxes
    # which may or may not screw up the r-tree intersection queries
    ray_bounding += np.array([-1, -1, -1, 1, 1, 1]) * buffer_dist

    return ray_bounding



# flake8: NOQA
"""
visual
-------------

Handle visual properties for meshes, including color and texture
"""

from .color import (
    ColorVisuals,
    random_color,
    to_rgba,
    DEFAULT_COLOR,
    interpolate,
    uv_to_color,
    uv_to_interpolated_color,
    linear_color_map,
)

from .texture import TextureVisuals
from .objects import create_visual, concatenate

from . import color
from . import texture
from . import objects
from . import material
from .. import resolvers


# explicitly list imports in __all__
# as otherwise flake8 gets mad
__all__ = [
    "color",
    "texture",
    "resolvers",
    "TextureVisuals",
    "ColorVisuals",
    "random_color",
    "to_rgba",
    "create_visual",
    "DEFAULT_COLOR",
    "interpolate",
    "linear_color_map",
    "uv_to_color",
    "uv_to_interpolated_color",
]



import numpy as np
from shapely import ops
from shapely.geometry import Polygon

from .. import bounds, geometry, graph, grouping
from ..constants import log
from ..constants import tol_path as tol
from ..transformations import transform_points
from ..typed import List, NDArray, Optional, float64
from .simplify import fit_circle_check
from .traversal import resample_path

try:
    import networkx as nx
except BaseException as E:
    # create a dummy module which will raise the ImportError
    # or other exception only when someone tries to use networkx
    from ..exceptions import ExceptionWrapper

    nx = ExceptionWrapper(E)
try:
    from rtree.index import Index
except BaseException as E:
    # create a dummy module which will raise the ImportError
    from ..exceptions import ExceptionWrapper

    Index = ExceptionWrapper(E)


def enclosure_tree(polygons: List[Polygon]):
    """
    Given a list of shapely polygons with only exteriors,
    find which curves represent the exterior shell or root curve
    and which represent holes which penetrate the exterior.

    This is done with an R-tree for rough overlap detection,
    and then exact polygon queries for a final result.

    Parameters
    -----------
    polygons : (n,) shapely.geometry.Polygon
       Polygons which only have exteriors and may overlap

    Returns
    -----------
    roots : (m,) int
        Index of polygons which are root
    contains : networkx.DiGraph
       Edges indicate a polygon is
       contained by another polygon
    """

    # get the bounds for every valid polygon
    bounds = {
        k: v
        for k, v in {
            i: getattr(polygon, "bounds", []) for i, polygon in enumerate(polygons)
        }.items()
        if len(v) == 4
    }

    # nodes are indexes in polygons
    contains = nx.DiGraph()
    # make sure we don't have orphaned polygon
    contains.add_nodes_from(bounds.keys())

    if len(bounds) > 0:
        # if there are no valid bounds tree creation will fail
        # and we won't be calling `tree.intersection` anywhere
        # we could return here but having multiple return paths
        # seems more dangerous than iterating through an empty graph
        tree = Index(zip(bounds.keys(), bounds.values(), [None] * len(bounds)))

    # loop through every polygon
    for i, b in bounds.items():
        # we first query for bounding box intersections from the R-tree
        for j in tree.intersection(b):
            # if we are checking a polygon against itself continue
            if i == j:
                continue
            # do a more accurate polygon in polygon test
            # for the enclosure tree information
            if polygons[i].contains(polygons[j]):
                contains.add_edge(i, j)
            elif polygons[j].contains(polygons[i]):
                contains.add_edge(j, i)

    # a root or exterior curve has an even number of parents
    # wrap in dict call to avoid networkx view
    degree = dict(contains.in_degree())
    # convert keys and values to numpy arrays
    indexes = np.array(list(degree.keys()))
    degrees = np.array(list(degree.values()))
    # roots are curves with an even inward degree (parent count)
    roots = indexes[(degrees % 2) == 0]
    # if there are multiple nested polygons split the graph
    # so the contains logic returns the individual polygons
    if len(degrees) > 0 and degrees.max() > 1:
        # collect new edges for graph
        edges = []
        # order the roots so they are sorted by degree
        roots = roots[np.argsort([degree[r] for r in roots])]
        # find edges of subgraph for each root and children
        for root in roots:
            children = indexes[degrees == degree[root] + 1]
            edges.extend(contains.subgraph(np.append(children, root)).edges())
        # stack edges into new directed graph
        contains = nx.from_edgelist(edges, nx.DiGraph())
        # if roots have no children add them anyway
        contains.add_nodes_from(roots)

    return roots, contains


def edges_to_polygons(edges, vertices):
    """
    Given an edge list of indices and associated vertices
    representing lines, generate a list of polygons.

    Parameters
    -----------
    edges : (n, 2) int
      Indexes of vertices which represent lines
    vertices : (m, 2) float
      Vertices in 2D space

    Returns
    ----------
    polygons : (p,) shapely.geometry.Polygon
      Polygon objects with interiors
    """

    assert isinstance(vertices, np.ndarray)

    # create closed polygon objects
    polygons = []
    # loop through a sequence of ordered traversals
    for dfs in graph.traversals(edges, mode="dfs"):
        try:
            # try to recover polygons before they are more complicated
            repaired = repair_invalid(Polygon(vertices[dfs]))
            # if it returned a multipolygon extend into a flat list
            if hasattr(repaired, "geoms"):
                polygons.extend(repaired.geoms)
            else:
                polygons.append(repaired)
        except ValueError:
            continue

    # if there is only one polygon, just return it
    if len(polygons) == 1:
        return polygons

    # find which polygons contain which other polygons
    roots, tree = enclosure_tree(polygons)

    # generate list of polygons with proper interiors
    complete = []
    for root in roots:
        interior = list(tree[root].keys())
        shell = polygons[root].exterior.coords
        holes = [polygons[i].exterior.coords for i in interior]
        complete.append(Polygon(shell=shell, holes=holes))
    return complete


def polygons_obb(polygons: List[Polygon]):
    """
    Find the OBBs for a list of shapely.geometry.Polygons
    """
    rectangles = [None] * len(polygons)
    transforms = [None] * len(polygons)
    for i, p in enumerate(polygons):
        transforms[i], rectangles[i] = polygon_obb(p)
    return np.array(transforms), np.array(rectangles)


def polygon_obb(polygon: Polygon):
    """
    Find the oriented bounding box of a Shapely polygon.

    The OBB is always aligned with an edge of the convex hull of the polygon.

    Parameters
    -------------
    polygons : shapely.geometry.Polygon
      Input geometry

    Returns
    -------------
    transform : (3, 3) float
      Transformation matrix
      which will move input polygon from its original position
      to the first quadrant where the AABB is the OBB
    extents : (2,) float
      Extents of transformed polygon
    """
    if hasattr(polygon, "exterior"):
        points = np.asanyarray(polygon.exterior.coords)
    elif isinstance(polygon, np.ndarray):
        points = polygon
    else:
        raise ValueError("polygon or points must be provided")

    transform, extents = bounds.oriented_bounds_2D(points)

    if tol.strict:
        moved = transform_points(points=points, matrix=transform)
        assert np.allclose(-extents / 2.0, moved.min(axis=0))
        assert np.allclose(extents / 2.0, moved.max(axis=0))

    return transform, extents


def transform_polygon(polygon, matrix):
    """
    Transform a polygon by a a 2D homogeneous transform.

    Parameters
    -------------
    polygon : shapely.geometry.Polygon
      2D polygon to be transformed.
    matrix  : (3, 3) float
      2D homogeneous transformation.

    Returns
    --------------
    result : shapely.geometry.Polygon
      Polygon transformed by matrix.
    """
    matrix = np.asanyarray(matrix, dtype=np.float64)

    if hasattr(polygon, "geoms"):
        result = [transform_polygon(p, t) for p, t in zip(polygon, matrix)]
        return result
    # transform the outer shell
    shell = transform_points(np.array(polygon.exterior.coords), matrix)[:, :2]
    # transform the interiors
    holes = [
        transform_points(np.array(i.coords), matrix)[:, :2] for i in polygon.interiors
    ]
    # create a new polygon with the result
    result = Polygon(shell=shell, holes=holes)
    return result


def polygon_bounds(polygon, matrix=None):
    """
    Get the transformed axis aligned bounding box of a
    shapely Polygon object.

    Parameters
    ------------
    polygon : shapely.geometry.Polygon
      Polygon pre-transform
    matrix : (3, 3) float or None.
      Homogeneous transform moving polygon in space

    Returns
    ------------
    bounds : (2, 2) float
      Axis aligned bounding box of transformed polygon.
    """
    if matrix is not None:
        assert matrix.shape == (3, 3)
        points = transform_points(points=np.array(polygon.exterior.coords), matrix=matrix)
    else:
        points = np.array(polygon.exterior.coords)

    bounds = np.array([points.min(axis=0), points.max(axis=0)])
    assert bounds.shape == (2, 2)
    return bounds


def plot(polygon=None, show=True, axes=None, **kwargs):
    """
    Plot a shapely polygon using matplotlib.

    Parameters
    ------------
    polygon : shapely.geometry.Polygon
      Polygon to be plotted
    show : bool
      If True will display immediately
    **kwargs
      Passed to plt.plot
    """
    import matplotlib.pyplot as plt

    def plot_single(single):
        axes.plot(*single.exterior.xy, **kwargs)
        for interior in single.interiors:
            axes.plot(*interior.xy, **kwargs)

    # make aspect ratio non-stupid
    if axes is None:
        axes = plt.axes()
    axes.set_aspect("equal", "datalim")

    if polygon.__class__.__name__ == "MultiPolygon":
        [plot_single(i) for i in polygon.geoms]
    elif hasattr(polygon, "__iter__"):
        [plot_single(i) for i in polygon]
    elif polygon is not None:
        plot_single(polygon)

    if show:
        plt.show()

    return axes


def resample_boundaries(polygon: Polygon, resolution: float, clip=None):
    """
    Return a version of a polygon with boundaries re-sampled
    to a specified resolution.

    Parameters
    -------------
    polygon : shapely.geometry.Polygon
      Source geometry
    resolution : float
      Desired distance between points on boundary
    clip : (2,) int
      Upper and lower bounds to clip
      number of samples to avoid exploding count

    Returns
    ------------
    kwargs : dict
     Keyword args for a Polygon constructor `Polygon(**kwargs)`
    """

    def resample_boundary(boundary):
        # add a polygon.exterior or polygon.interior to
        # the deque after resampling based on our resolution
        count = boundary.length / resolution
        count = int(np.clip(count, *clip))
        return resample_path(boundary.coords, count=count)

    if clip is None:
        clip = [8, 200]
    # create a sequence of [(n,2)] points
    kwargs = {"shell": resample_boundary(polygon.exterior), "holes": []}
    for interior in polygon.interiors:
        kwargs["holes"].append(resample_boundary(interior))

    return kwargs


def stack_boundaries(boundaries):
    """
    Stack the boundaries of a polygon into a single
    (n, 2) list of vertices.

    Parameters
    ------------
    boundaries : dict
      With keys 'shell', 'holes'

    Returns
    ------------
    stacked : (n, 2) float
      Stacked vertices
    """
    if len(boundaries["holes"]) == 0:
        return boundaries["shell"]
    result = np.vstack((boundaries["shell"], np.vstack(boundaries["holes"])))
    return result


def medial_axis(polygon: Polygon, resolution: Optional[float] = None, clip=None):
    """
    Given a shapely polygon, find the approximate medial axis
    using a voronoi diagram of evenly spaced points on the
    boundary of the polygon.

    Parameters
    ----------
    polygon : shapely.geometry.Polygon
      The source geometry
    resolution : float
      Distance between each sample on the polygon boundary
    clip : None, or (2,) int
      Clip sample count to min of clip[0] and max of clip[1]

    Returns
    ----------
    edges : (n, 2) int
      Vertex indices representing line segments
      on the polygon's medial axis
    vertices : (m, 2) float
      Vertex positions in space
    """
    # a circle will have a single point medial axis
    if len(polygon.interiors) == 0:
        # what is the approximate scale of the polygon
        scale = np.reshape(polygon.bounds, (2, 2)).ptp(axis=0).max()
        # a (center, radius, error) tuple
        fit = fit_circle_check(polygon.exterior.coords, scale=scale)
        # is this polygon in fact a circle
        if fit is not None:
            # return an edge that has the center as the midpoint
            epsilon = np.clip(fit["radius"] / 500, 1e-5, np.inf)
            vertices = np.array(
                [fit["center"] + [0, epsilon], fit["center"] - [0, epsilon]],
                dtype=np.float64,
            )
            # return a single edge to avoid consumers needing to special case
            edges = np.array([[0, 1]], dtype=np.int64)
            return edges, vertices

    from scipy.spatial import Voronoi
    from shapely import vectorized

    if resolution is None:
        resolution = np.reshape(polygon.bounds, (2, 2)).ptp(axis=0).max() / 100

    # get evenly spaced points on the polygons boundaries
    samples = resample_boundaries(polygon=polygon, resolution=resolution, clip=clip)
    # stack the boundary into a (m,2) float array
    samples = stack_boundaries(samples)
    # create the voronoi diagram on 2D points
    voronoi = Voronoi(samples)
    # which voronoi vertices are contained inside the polygon
    contains = vectorized.contains(polygon, *voronoi.vertices.T)
    # ridge vertices of -1 are outside, make sure they are False
    contains = np.append(contains, False)
    # make sure ridge vertices is numpy array
    ridge = np.asanyarray(voronoi.ridge_vertices, dtype=np.int64)
    # only take ridges where every vertex is contained
    edges = ridge[contains[ridge].all(axis=1)]

    # now we need to remove uncontained vertices
    contained = np.unique(edges)
    mask = np.zeros(len(voronoi.vertices), dtype=np.int64)
    mask[contained] = np.arange(len(contained))

    # mask voronoi vertices
    vertices = voronoi.vertices[contained]
    # re-index edges
    edges_final = mask[edges]

    if tol.strict:
        # make sure we didn't screw up indexes
        assert (vertices[edges_final] - voronoi.vertices[edges]).ptp() < 1e-5

    return edges_final, vertices


def identifier(polygon: Polygon) -> NDArray[float64]:
    """
    Return a vector containing values representative of
    a particular polygon.

    Parameters
    ---------
    polygon : shapely.geometry.Polygon
      Input geometry

    Returns
    ---------
    identifier : (8,) float
      Values which should be unique for this polygon.
    """
    result = [
        len(polygon.interiors),
        polygon.convex_hull.area,
        polygon.convex_hull.length,
        polygon.area,
        polygon.length,
        polygon.exterior.length,
    ]
    # include the principal second moments of inertia of the polygon
    # this is invariant to rotation and translation
    _, principal, _, _ = second_moments(polygon, return_centered=True)
    result.extend(principal)

    return np.array(result, dtype=np.float64)


def random_polygon(segments=8, radius=1.0):
    """
    Generate a random polygon with a maximum number of sides and approximate radius.

    Parameters
    ---------
    segments : int
      The maximum number of sides the random polygon will have
    radius : float
      The approximate radius of the polygon desired

    Returns
    ---------
    polygon : shapely.geometry.Polygon
      Geometry object with random exterior and no interiors.
    """
    angles = np.sort(np.cumsum(np.random.random(segments) * np.pi * 2) % (np.pi * 2))
    radii = np.random.random(segments) * radius

    points = np.column_stack((np.cos(angles), np.sin(angles))) * radii.reshape((-1, 1))
    points = np.vstack((points, points[0]))
    polygon = Polygon(points).buffer(0.0)
    if hasattr(polygon, "geoms"):
        return polygon.geoms[0]
    return polygon


def polygon_scale(polygon):
    """
    For a Polygon object return the diagonal length of the AABB.

    Parameters
    ------------
    polygon : shapely.geometry.Polygon
      Source geometry

    Returns
    ------------
    scale : float
      Length of AABB diagonal
    """
    extents = np.reshape(polygon.bounds, (2, 2)).ptp(axis=0)
    scale = (extents**2).sum() ** 0.5

    return scale


def paths_to_polygons(paths, scale=None):
    """
    Given a sequence of connected points turn them into
    valid shapely Polygon objects.

    Parameters
    -----------
    paths : (n,) sequence
      Of (m, 2) float closed paths
    scale : float
      Approximate scale of drawing for precision

    Returns
    -----------
    polys : (p,) list
      Filled with Polygon or None

    """
    polygons = [None] * len(paths)
    for i, path in enumerate(paths):
        if len(path) < 4:
            # since the first and last vertices are identical in
            # a closed loop a 4 vertex path is the minimum for
            # non-zero area
            continue
        try:
            polygons[i] = repair_invalid(Polygon(path), scale)
        except ValueError:
            # raised if a polygon is unrecoverable
            continue
        except BaseException:
            log.error("unrecoverable polygon", exc_info=True)
    polygons = np.array(polygons)
    return polygons


def sample(polygon, count, factor=1.5, max_iter=10):
    """
    Use rejection sampling to generate random points inside a
    polygon.

    Parameters
    -----------
    polygon : shapely.geometry.Polygon
      Polygon that will contain points
    count : int
      Number of points to return
    factor : float
      How many points to test per loop
    max_iter : int
      Maximum number of intersection checks is:
      > count * factor * max_iter

    Returns
    -----------
    hit : (n, 2) float
      Random points inside polygon
      where n <= count
    """
    # do batch point-in-polygon queries
    from shapely import vectorized

    # get size of bounding box
    bounds = np.reshape(polygon.bounds, (2, 2))
    extents = bounds.ptp(axis=0)

    # how many points to check per loop iteration
    per_loop = int(count * factor)

    # start with some rejection sampling
    points = bounds[0] + extents * np.random.random((per_loop, 2))
    # do the point in polygon test and append resulting hits
    mask = vectorized.contains(polygon, *points.T)
    hit = [points[mask]]
    hit_count = len(hit[0])
    # if our first non-looping check got enough samples exit
    if hit_count >= count:
        return hit[0][:count]

    # if we have to do iterations loop here slowly
    for _ in range(max_iter):
        # generate points inside polygons AABB
        points = np.random.random((per_loop, 2))
        points = (points * extents) + bounds[0]
        # do the point in polygon test and append resulting hits
        mask = vectorized.contains(polygon, *points.T)
        hit.append(points[mask])
        # keep track of how many points we've collected
        hit_count += len(hit[-1])
        # if we have enough points exit the loop
        if hit_count > count:
            break

    # stack the hits into an (n,2) array and truncate
    hit = np.vstack(hit)[:count]

    return hit


def repair_invalid(polygon, scale=None, rtol=0.5):
    """
    Given a shapely.geometry.Polygon, attempt to return a
    valid version of the polygon through buffering tricks.

    Parameters
    -----------
    polygon : shapely.geometry.Polygon
      Source geometry
    rtol : float
      How close does a perimeter have to be
    scale : float or None
      For numerical precision reference

    Returns
    ----------
    repaired : shapely.geometry.Polygon
      Repaired polygon

    Raises
    ----------
    ValueError
      If polygon can't be repaired
    """
    if hasattr(polygon, "is_valid") and polygon.is_valid:
        return polygon

    # basic repair involves buffering the polygon outwards
    # this will fix a subset of problems.
    basic = polygon.buffer(tol.zero)
    # if it returned multiple polygons check the largest
    if hasattr(basic, "geoms"):
        basic = basic.geoms[np.argmax([i.area for i in basic.geoms])]

    # check perimeter of result against original perimeter
    if basic.is_valid and np.isclose(basic.length, polygon.length, rtol=rtol):
        return basic

    if scale is None:
        distance = 0.002 * np.reshape(polygon.bounds, (2, 2)).ptp(axis=0).mean()
    else:
        distance = 0.002 * scale

    # if there are no interiors, we can work with just the exterior
    # ring, which is often more reliable
    if len(polygon.interiors) == 0:
        # try buffering the exterior of the polygon
        # the interior will be offset by -tol.buffer
        rings = polygon.exterior.buffer(distance).interiors
        if len(rings) == 1:
            # reconstruct a single polygon from the interior ring
            recon = Polygon(shell=rings[0]).buffer(distance)
            # check perimeter of result against original perimeter
            if recon.is_valid and np.isclose(recon.length, polygon.length, rtol=rtol):
                return recon

        # try de-deuplicating the outside ring
        points = np.array(polygon.exterior.coords)
        # remove any segments shorter than tol.merge
        # this is a little risky as if it was discretized more
        # finely than 1-e8 it may remove detail
        unique = np.append(True, (np.diff(points, axis=0) ** 2).sum(axis=1) ** 0.5 > 1e-8)
        # make a new polygon with result
        dedupe = Polygon(shell=points[unique])
        # check result
        if dedupe.is_valid and np.isclose(dedupe.length, polygon.length, rtol=rtol):
            return dedupe

    # buffer and unbuffer the whole polygon
    buffered = polygon.buffer(distance).buffer(-distance)
    # if it returned multiple polygons check the largest
    if hasattr(buffered, "geoms"):
        areas = np.array([b.area for b in buffered.geoms])
        return buffered.geoms[areas.argmax()]

    # check perimeter of result against original perimeter
    if buffered.is_valid and np.isclose(buffered.length, polygon.length, rtol=rtol):
        log.debug("Recovered invalid polygon through double buffering")
        return buffered

    raise ValueError("unable to recover polygon!")


def projected(
    mesh,
    normal,
    origin=None,
    ignore_sign=True,
    rpad=1e-5,
    apad=None,
    tol_dot=0.01,
    max_regions=200,
):
    """
    Project a mesh onto a plane and then extract the polygon
    that outlines the mesh projection on that plane.

    Note that this will ignore back-faces, which is only
    relevant if the source mesh isn't watertight.

    Also padding: this generates a result by unioning the
    polygons of multiple connected regions, which requires
    the polygons be padded by a distance so that a polygon
    union produces a single coherent result. This distance
    is calculated as: `apad + (rpad * scale)`

    Parameters
    ----------
    mesh : trimesh.Trimesh
      Source geometry
    check : bool
      If True make sure is flat
    normal : (3,) float
      Normal to extract flat pattern along
    origin : None or (3,) float
      Origin of plane to project mesh onto
    ignore_sign : bool
      Allow a projection from the normal vector in
      either direction: this provides a substantial speedup
      on watertight meshes where the direction is irrelevant
      but if you have a triangle soup and want to discard
      backfaces you should set this to False.
    rpad : float
      Proportion to pad polygons by before unioning
      and then de-padding result by to avoid zero-width gaps.
    apad : float
      Absolute padding to pad polygons by before unioning
      and then de-padding result by to avoid zero-width gaps.
    tol_dot : float
      Tolerance for discarding on-edge triangles.
    max_regions : int
      Raise an exception if the mesh has more than this
      number of disconnected regions to fail quickly before
      unioning.

    Returns
    ----------
    projected : shapely.geometry.Polygon or None
      Outline of source mesh

    Raises
    ---------
    ValueError
      If max_regions is exceeded
    """
    # make sure normal is a unitized copy
    normal = np.array(normal, dtype=np.float64)
    normal /= np.linalg.norm(normal)

    # the projection of each face normal onto facet normal
    dot_face = np.dot(normal, mesh.face_normals.T)
    if ignore_sign:
        # for watertight mesh speed up projection by handling side with less faces
        # check if face lies on front or back of normal
        front = dot_face > tol_dot
        back = dot_face < -tol_dot
        # divide the mesh into front facing section and back facing parts
        # and discard the faces perpendicular to the axis.
        # since we are doing a unary_union later we can use the front *or*
        # the back so we use which ever one has fewer triangles
        # we want the largest nonzero group
        count = np.array([front.sum(), back.sum()])
        if count.min() == 0:
            # if one of the sides has zero faces we need the other
            pick = count.argmax()
        else:
            # otherwise use the normal direction with the fewest faces
            pick = count.argmin()
        # use the picked side
        side = [front, back][pick]
    else:
        # if explicitly asked to care about the sign
        # only handle the front side of normal
        side = dot_face > tol_dot

    # subset the adjacency pairs to ones which have both faces included
    # on the side we are currently looking at
    adjacency_check = side[mesh.face_adjacency].all(axis=1)
    adjacency = mesh.face_adjacency[adjacency_check]

    # a sequence of face indexes that are connected
    face_groups = graph.connected_components(adjacency, nodes=np.nonzero(side)[0])

    # if something is goofy we may end up with thousands of
    # regions that do nothing except hang for an hour then segfault
    if len(face_groups) > max_regions:
        raise ValueError("too many disconnected groups!")

    # reshape edges into shape length of faces for indexing
    edges = mesh.edges_sorted.reshape((-1, 6))
    # transform from the mesh frame in 3D to the XY plane
    to_2D = geometry.plane_transform(origin=origin, normal=normal)
    # transform mesh vertices to 2D and clip the zero Z
    vertices_2D = transform_points(mesh.vertices, to_2D)[:, :2]

    polygons = []
    for faces in face_groups:
        # index edges by face then shape back to individual edges
        edge = edges[faces].reshape((-1, 2))
        # edges that occur only once are on the boundary
        group = grouping.group_rows(edge, require_count=1)
        # turn each region into polygons
        polygons.extend(edges_to_polygons(edges=edge[group], vertices=vertices_2D))

    padding = 0.0
    if apad is not None:
        # set padding by absolute value
        padding += float(apad)
    if rpad is not None:
        # get the 2D scale as the longest side of the AABB
        scale = vertices_2D.ptp(axis=0).max()
        # apply the scale-relative padding
        padding += float(rpad) * scale

    # some types of errors will lead to a bajillion disconnected
    # regions and the union will take forever to fail
    # so exit here early
    if len(polygons) > max_regions:
        raise ValueError("too many disconnected groups!")

    # if there is only one region we don't need to run a union
    elif len(polygons) == 1:
        return polygons[0]
    elif len(polygons) == 0:
        return None
    else:
        # inflate each polygon before unioning to remove zero-size
        # gaps then deflate the result after unioning by the same amount
        # note the following provides a 25% speedup but needs
        # more testing to see if it deflates to a decent looking
        # result:
        # polygon = ops.unary_union(
        #    [p.buffer(padding,
        #              join_style=2,
        #              mitre_limit=1.5)
        #     for p in polygons]).buffer(-padding)
        polygon = ops.unary_union([p.buffer(padding) for p in polygons]).buffer(-padding)
    return polygon


def second_moments(polygon: Polygon, return_centered=False):
    """
    Calculate the second moments of area of a polygon
    from the boundary.

    Parameters
    ------------
    polygon : shapely.geometry.Polygon
      Closed polygon.
    return_centered : bool
      Get second moments for a frame with origin at the centroid
      and perform a principal axis transformation.

    Returns
    ----------
    moments : (3,) float
      The values of `[Ixx, Iyy, Ixy]`
    principal_moments : (2,) float
      Principal second moments of inertia: `[Imax, Imin]`
      Only returned if `centered`.
    alpha : float
      Angle by which the polygon needs to be rotated, so the
      principal axis align with the X and Y axis.
      Only returned if `centered`.
    transform : (3, 3) float
      Transformation matrix which rotates the polygon by alpha.
      Only returned if `centered`.
    """

    transform = np.eye(3)
    if return_centered:
        # calculate centroid and move polygon
        transform[:2, 2] = -np.array(polygon.centroid.coords)
        polygon = transform_polygon(polygon, transform)

    # start with the exterior
    coords = np.array(polygon.exterior.coords)
    # shorthand the coordinates
    x1, y1 = np.vstack((coords[-1], coords[:-1])).T
    x2, y2 = coords.T
    # do vectorized operations
    v = x1 * y2 - x2 * y1
    Ixx = np.sum(v * (y1 * y1 + y1 * y2 + y2 * y2)) / 12.0
    Iyy = np.sum(v * (x1 * x1 + x1 * x2 + x2 * x2)) / 12.0
    Ixy = np.sum(v * (x1 * y2 + 2 * x1 * y1 + 2 * x2 * y2 + x2 * y1)) / 24.0

    for interior in polygon.interiors:
        coords = np.array(interior.coords)
        # shorthand the coordinates
        x1, y1 = np.vstack((coords[-1], coords[:-1])).T
        x2, y2 = coords.T
        # do vectorized operations
        v = x1 * y2 - x2 * y1
        Ixx -= np.sum(v * (y1 * y1 + y1 * y2 + y2 * y2)) / 12.0
        Iyy -= np.sum(v * (x1 * x1 + x1 * x2 + x2 * x2)) / 12.0
        Ixy -= np.sum(v * (x1 * y2 + 2 * x1 * y1 + 2 * x2 * y2 + x2 * y1)) / 24.0

    moments = [Ixx, Iyy, Ixy]

    if not return_centered:
        return moments

    # get the principal moments
    root = np.sqrt(((Iyy - Ixx) / 2.0) ** 2 + Ixy**2)
    Imax = (Ixx + Iyy) / 2.0 + root
    Imin = (Ixx + Iyy) / 2.0 - root
    principal_moments = [Imax, Imin]

    # do the principal axis transform
    if np.isclose(Ixy, 0.0, atol=1e-12):
        alpha = 0
    elif np.isclose(Ixx, Iyy):
        # prevent division by 0
        alpha = 0.25 * np.pi
    else:
        alpha = 0.5 * np.arctan(2.0 * Ixy / (Ixx - Iyy))

    # construct transformation matrix
    cos_alpha = np.cos(alpha)
    sin_alpha = np.sin(alpha)

    transform[0, 0] = cos_alpha
    transform[1, 1] = cos_alpha
    transform[0, 1] = -sin_alpha
    transform[1, 0] = sin_alpha

    return moments, principal_moments, alpha, transform



"""
interval.py
--------------

Deal with 1D intervals which are defined by:
  [start position, end position]
"""

import numpy as np

from .typed import NDArray, float64


def intersection(a: NDArray[float64], b: NDArray[float64]) -> NDArray[float64]:
    """
    Given pairs of ranges merge them in to
    one range if they overlap.

    Parameters
    --------------
    a : (2, ) or (n, 2)
      Start and end of a 1D interval
    b : (2, ) float
      Start and end of a 1D interval

    Returns
    --------------
    inter : (2, ) or (2, 2) float
      The unioned range from the two inputs,
      if not `inter.ptp(axis=1)` will be zero.
    """
    a = np.array(a, dtype=np.float64)
    b = np.array(b, dtype=np.float64)

    # convert to vectorized form
    is_1D = a.shape == (2,)
    a = a.reshape((-1, 2))
    b = b.reshape((-1, 2))

    # make sure they're min-max
    a.sort(axis=1)
    b.sort(axis=1)
    a_low, a_high = a.T
    b_low, b_high = b.T

    # do the checks
    check = np.logical_not(np.logical_or(b_low >= a_high, a_low >= b_high))
    overlap = np.zeros(a.shape, dtype=np.float64)
    overlap[check] = np.column_stack(
        (
            np.array([a_low[check], b_low[check]]).max(axis=0),
            np.array([a_high[check], b_high[check]]).min(axis=0),
        )
    )

    if is_1D:
        return overlap[0]

    return overlap


def union(intervals: NDArray[float64], sort: bool = True) -> NDArray[float64]:
    """
    For array of multiple intervals union them all into
    the subset of intervals.

    For example:
    `intervals = [[1,2], [2,3]] -> [[1, 3]]`
    `intervals = [[1,2], [2.5,3]] -> [[1, 2], [2.5, 3]]`


    Parameters
    ------------
    intervals : (n, 2)
      Pairs of `(min, max)` values.
    sort
      If the array is already ordered into (min, max) pairs
      and then pairs sorted by minimum value you can skip the
      sorting in this function.

    Returns
    ----------
    unioned : (m, 2)
      New intervals where `m <= n`
    """
    if len(intervals) == 0:
        return np.empty(0)

    # if the intervals have not been pre-sorted we should apply our sorting logic
    # you would only skip this if you are subsetting a larger list elsewhere.
    if sort:
        # copy inputs and make sure they are (min, max) pairs
        intervals = np.sort(intervals, axis=1)
        # order them by lowest starting point
        intervals = intervals[intervals[:, 0].argsort()]

    # we know we will have at least one interval
    unions = [intervals[0]]

    for begin, end in intervals[1:]:
        if unions[-1][1] >= begin:
            #
            unions[-1][1] = max(unions[-1][1], end)
        else:
            unions.append([begin, end])

    return np.array(unions)



from dataclasses import dataclass

import numpy as np

from .. import util
from ..constants import log
from ..constants import res_path as res
from ..constants import tol_path as tol
from ..typed import ArrayLike, NDArray, Optional, float64

# floating point zero
_TOL_ZERO = 1e-12


@dataclass
class ArcInfo:
    # What is the radius of the circular arc?
    radius: float

    # what is the center of the circular arc
    # it is either 2D or 3D depending on input.
    center: NDArray[float64]

    # what is the 3D normal vector of the plane the arc lies on
    normal: Optional[NDArray[float64]] = None

    # what is the starting and ending angle of the arc.
    angles: Optional[NDArray[float64]] = None

    # what is the angular span of this circular arc.
    span: Optional[float] = None

    def __getitem__(self, item):
        # add for backwards compatibility
        return getattr(self, item)


def arc_center(
    points: ArrayLike, return_normal: bool = True, return_angle: bool = True
) -> ArcInfo:
    """
    Given three points on a 2D or 3D arc find the center,
    radius, normal, and angular span.

    Parameters
    ---------
    points : (3, dimension) float
      Points in space, where dimension is either 2 or 3
    return_normal : bool
      If True calculate the 3D normal unit vector
    return_angle : bool
      If True calculate the start and stop angle and span

    Returns
    ---------
    info
      Arc center, radius, and other information.
    """
    points = np.asanyarray(points, dtype=np.float64)

    # get the non-unit vectors of the three points
    vectors = points[[2, 0, 1]] - points[[1, 2, 0]]
    # we need both the squared row sum and the non-squared
    abc2 = np.dot(vectors**2, [1] * points.shape[1])
    # same as np.linalg.norm(vectors, axis=1)
    abc = np.sqrt(abc2)

    # perform radius calculation scaled to shortest edge
    # to avoid precision issues with small or large arcs
    scale = abc.min()
    # get the edge lengths scaled to the smallest
    edges = abc / scale
    # half the total length of the edges
    half = edges.sum() / 2.0
    # check the denominator for the radius calculation
    denom = half * np.prod(half - edges)
    if denom < tol.merge:
        raise ValueError("arc is colinear!")
    # find the radius and scale back after the operation
    radius = scale * ((np.prod(edges) / 4.0) / np.sqrt(denom))

    # use a barycentric approach to get the center
    ba2 = (abc2[[1, 2, 0, 0, 2, 1, 0, 1, 2]] * [1, 1, -1, 1, 1, -1, 1, 1, -1]).reshape(
        (3, 3)
    ).sum(axis=1) * abc2
    center = points.T.dot(ba2) / ba2.sum()

    if tol.strict:
        # all points should be at the calculated radius from center
        assert util.allclose(np.linalg.norm(points - center, axis=1), radius)

    # start with initial results
    result = {"center": center, "radius": radius}
    if return_normal:
        if points.shape == (3, 2):
            # for 2D arcs still use the cross product so that
            # the sign of the normal vector is consistent
            result["normal"] = util.unitize(
                np.cross(np.append(-vectors[1], 0), np.append(vectors[2], 0))
            )
        else:
            # otherwise just take the cross product
            result["normal"] = util.unitize(np.cross(-vectors[1], vectors[2]))

    if return_angle:
        # vectors from points on arc to center point
        vector = util.unitize(points - center)
        edge_direction = np.diff(points, axis=0)
        # find the angle between the first and last vector
        dot = np.dot(*vector[[0, 2]])
        if dot < (_TOL_ZERO - 1):
            angle = np.pi
        elif dot > 1 - _TOL_ZERO:
            angle = 0.0
        else:
            angle = np.arccos(dot)
        # if the angle is nonzero and vectors are opposite direction
        # it means we have a long arc rather than the short path
        if abs(angle) > _TOL_ZERO and np.dot(*edge_direction) < 0.0:
            angle = (np.pi * 2) - angle
        # convoluted angle logic
        angles = np.arctan2(*vector[:, :2].T[::-1]) + np.pi * 2
        angles_sorted = np.sort(angles[[0, 2]])
        reverse = angles_sorted[0] < angles[1] < angles_sorted[1]
        angles_sorted = angles_sorted[:: (1 - int(not reverse) * 2)]
        result["angles"] = angles_sorted
        result["span"] = angle

    return ArcInfo(**result)


def discretize_arc(points, close=False, scale=1.0):
    """
    Returns a version of a three point arc consisting of
    line segments.

    Parameters
    ---------
    points : (3, d) float
      Points on the arc where d in [2,3]
    close :  boolean
      If True close the arc into a circle
    scale : float
      What is the approximate overall drawing scale
      Used to establish order of magnitude for precision

    Returns
    ---------
    discrete : (m, d) float
      Connected points in space
    """
    # make sure points are (n, 3)
    points, is_2D = util.stack_3D(points, return_2D=True)
    # find the center of the points
    try:
        # try to find the center from the arc points
        center_info = arc_center(points)
    except BaseException:
        # if we hit an exception return a very bad but
        # technically correct discretization of the arc
        if is_2D:
            return points[:, :2]
        return points

    center, R, N, angle = (
        center_info.center,
        center_info.radius,
        center_info.normal,
        center_info.span,
    )

    # if requested, close arc into a circle
    if close:
        angle = np.pi * 2

    # the number of facets, based on the angle criteria
    count_a = angle / res.seg_angle
    count_l = (R * angle) / (res.seg_frac * scale)

    # figure out the number of line segments
    count = np.max([count_a, count_l])
    # force at LEAST 4 points for the arc
    # otherwise the endpoints will diverge
    count = np.clip(count, 4, np.inf)
    count = int(np.ceil(count))

    V1 = util.unitize(points[0] - center)
    V2 = util.unitize(np.cross(-N, V1))
    t = np.linspace(0, angle, count)

    discrete = np.tile(center, (count, 1))
    discrete += R * np.cos(t).reshape((-1, 1)) * V1
    discrete += R * np.sin(t).reshape((-1, 1)) * V2

    # do an in-process check to make sure result endpoints
    # match the endpoints of the source arc
    if tol.strict and not close:
        arc_dist = util.row_norm(points[[0, -1]] - discrete[[0, -1]])
        arc_ok = (arc_dist < tol.merge).all()
        if not arc_ok:
            log.warning(
                "failed to discretize arc (endpoint_distance=%s R=%s)", str(arc_dist), R
            )
            log.warning("Failed arc points: %s", str(points))
            raise ValueError("Arc endpoints diverging!")
    discrete = discrete[:, : (3 - is_2D)]

    return discrete


def to_threepoint(center, radius, angles=None):
    """
    For 2D arcs, given a center and radius convert them to three
    points on the arc.

    Parameters
    -----------
    center : (2,) float
      Center point on the plane
    radius : float
      Radius of arc
    angles : (2,) float
      Angles in radians for start and end angle
      if not specified, will default to (0.0, pi)

    Returns
    ----------
    three : (3, 2) float
      Arc control points
    """
    # if no angles provided assume we want a half circle
    if angles is None:
        angles = [0.0, np.pi]
    # force angles to float64
    angles = np.asanyarray(angles, dtype=np.float64)
    if angles.shape != (2,):
        raise ValueError("angles must be (2,)!")
    # provide the wrap around
    if angles[1] < angles[0]:
        angles[1] += np.pi * 2

    center = np.asanyarray(center, dtype=np.float64)
    if center.shape != (2,):
        raise ValueError("only valid on 2D arcs!")

    # turn the angles of [start, end]
    # into [start, middle, end]
    angles = np.array([angles[0], angles.mean(), angles[1]], dtype=np.float64)
    # turn angles into (3, 2) points
    three = (np.column_stack((np.cos(angles), np.sin(angles))) * radius) + center

    return three



try:
    from . import generic as g
except BaseException:
    import generic as g


class URDFTest(g.unittest.TestCase):
    """
    Test to make sure we can convert trimesh objects
    to pyglet objects for rendering.
    """

    def test_export(self):
        # not imported to trimesh by __init__
        # so we don't have a pyglet import attempt
        # unless we explicitly ask for it (like during
        # a viewer show() call)
        from trimesh.exchange import urdf

        mesh = g.get_mesh("featuretype.STL")

        out_dir = g.tempfile.mkdtemp()

        urdf.export_urdf(mesh=mesh, directory=out_dir)

        # make sure it wrote something
        files = g.os.listdir(out_dir)
        assert len(files) > 0

        # remove temporary directory
        g.shutil.rmtree(out_dir)


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    g.unittest.main()



try:
    from . import generic as g
except BaseException:
    import generic as g

TEST_DIM = (100000, 3)


class CacheTest(g.unittest.TestCase):
    def test_track(self):
        """
        Check to make sure our fancy caching system only changes
        hashes when data actually changes.
        """

        original = g.trimesh.caching.hash_fast
        options = [
            g.trimesh.caching.hash_fast,
            g.trimesh.caching.hash_fallback,
            g.trimesh.caching.sha256,
        ]

        for option in options:
            g.log.info(f"testing hash function: {option.__name__}")
            g.trimesh.caching.hash_fast = option

            # generate test data and perform numpy operations
            a = g.trimesh.caching.tracked_array(g.random(TEST_DIM))
            modified = [hash(a)]
            a[0][0] = 10
            modified.append(hash(a))
            a[0][0] += 0.1
            modified.append(hash(a))
            a[:10] = g.np.fliplr(a[:10])
            modified.append(hash(a))
            a[1] = 5
            modified.append(hash(a))
            a[2:] = 2
            modified.append(hash(a))
            # these operations altered data and
            # the hash SHOULD have changed
            modified = g.np.array(modified, dtype=g.np.int64)
            assert (g.np.diff(modified) != 0).all()

            # now do slice operations which don't alter data
            modified = []
            modified.append(hash(a))
            b = a[[0, 1, 2]]  # NOQA
            modified.append(hash(a))
            c = a[1:]  # NOQA
            modified.append(hash(a))
            # double slice brah
            a = a[::-1][::-1]
            modified.append(hash(a))
            # these operations should have been cosmetic and
            # the hash should NOT have changed
            modified = g.np.array(modified, dtype=g.np.int64)
            assert (g.np.diff(modified, axis=0) == 0).all()

            # now change stuff and see if checksums change
            a = g.trimesh.caching.tracked_array([0, 0, 4])
            modified = []

            modified.append(hash(a))
            a += 10
            modified.append(hash(a))
            # assign some new data
            a = g.trimesh.caching.tracked_array([0.125, 115.32444, 4], dtype=g.np.float64)

            modified.append(hash(a))
            a += [10, 0, 0]
            modified.append(hash(a))
            a *= 10
            modified.append(hash(a))
            # itruediv rather than idiv
            a /= 2.0
            modified.append(hash(a))
            # idiv
            a /= 2.123
            modified.append(hash(a))
            a -= 1.0
            modified.append(hash(a))
            # in place floor division :|
            a //= 2
            modified.append(hash(a))
            # in-place sort
            a.sort()
            modified.append(hash(a))

            # these operations altered data and
            # the hash SHOULD have changed
            modified = g.np.array(modified, dtype=g.np.int64)
            assert (g.np.diff(modified) != 0).all()

        # reset our patched hash function
        g.trimesh.caching.hash_fast = original

    def test_contiguous(self):
        a = g.random((100, 3))
        t = g.trimesh.caching.tracked_array(a)

        original = g.trimesh.caching.hash_fast
        options = [
            g.trimesh.caching.hash_fast,
            g.trimesh.caching.hash_fallback,
            g.trimesh.caching.sha256,
        ]

        for option in options:
            g.log.info(f"testing hash function: {option.__name__}")
            g.trimesh.caching.hash_fast = option
            # hashing will fail on non- contiguous arrays
            # make sure our utility function has handled this
            assert hash(t) != hash(t[::-1])

        # reset our patched hash function
        g.trimesh.caching.hash_fast = original

    def test_mutable(self):
        """
        Run some simple tests on mutable DataStore objects
        """
        d = g.trimesh.caching.DataStore()

        d["hi"] = g.random(100)
        hash_initial = hash(d)
        # mutate internal data
        d["hi"][0] += 1
        assert hash(d) != hash_initial

        # should be mutable by default
        assert d.mutable
        # set data to immutable
        d.mutable = False

        try:
            d["hi"][1] += 1
        except ValueError:
            # should be raised when array is marked as read only
            return
        # we shouldn't make it past the try-except
        raise ValueError("mutating data worked when it shouldn't!")

    def test_transform(self):
        """
        apply_transform tries to not dump the full cache
        """
        m = g.get_mesh("featuretype.STL")
        # should populate edges_face
        e_len = len(m.edges)
        # should maintain required properties
        m.apply_transform(g.transforms[1])
        # should still be in the cache
        assert len(m.edges_face) == e_len

    def test_simple_collision(self):
        faces1 = g.np.array(
            [
                0,
                1,
                2,
                0,
                3,
                1,
                0,
                2,
                4,
                0,
                4,
                5,
                5,
                6,
                3,
                5,
                3,
                0,
                7,
                1,
                3,
                7,
                3,
                6,
                4,
                2,
                1,
                4,
                1,
                7,
                5,
                4,
                7,
                6,
                5,
                7,
            ],
            dtype=g.np.int64,
        ).reshape(-1, 3)
        faces2 = g.np.array(
            [
                0,
                1,
                2,
                0,
                3,
                1,
                2,
                4,
                0,
                5,
                4,
                2,
                6,
                3,
                0,
                6,
                0,
                4,
                6,
                1,
                3,
                6,
                7,
                1,
                2,
                7,
                5,
                2,
                1,
                7,
                4,
                5,
                7,
                6,
                4,
                7,
            ],
            dtype=g.np.int64,
        ).reshape(-1, 3)
        hash_fast = g.trimesh.caching.hash_fast
        assert hash_fast(faces1) != hash_fast(faces2)

    def test_scalar(self):
        # check to make sure we're doing the same
        # thing as numpy with our subclass
        o = g.np.arange(10)
        a = g.trimesh.caching.tracked_array(o)
        # tracked array should be different type
        # but with the same values (i.e. subclass)
        assert not isinstance(o, type(a))
        assert isinstance(a, g.np.ndarray)
        assert g.np.allclose(a, o)

        # this should be a scalar and subclasses
        # for some reason return a different type?
        assert isinstance(a.max(), type(o.max()))

    def test_method_combinations(self):
        # run combinatorial guesses of arguments on every method
        # of the subclassed `ndarray` and see if we can get trackedarray
        # to return incorrect hashes. Hopefully this catches new methods

        # only run this test on python 3+ as they changed tobytes
        if not g.PY3:
            return

        import itertools
        import warnings

        import numpy as np

        from trimesh.caching import tracked_array

        dim = (100, 3)

        # generate a bunch of arguments for every function of an `ndarray` so
        # we can see if the functions mutate
        flat = [
            2.3,
            1,
            10,
            4.2,
            [3, -1],
            {"shape": 10},
            np.int64,
            np.float64,
            True,
            True,
            False,
            False,
            g.random(dim),
            g.random(dim[::1]),
            "shape",
        ]

        # start with no arguments
        attempts = [()]
        # add a single argument from our guesses
        attempts.extend([(A,) for A in flat])
        # add 2 and 3 length permutations of our guesses
        attempts.extend([tuple(G) for G in itertools.product(flat, repeat=2)])

        # adding 3-length permutations makes this test 10x slower but if you
        # are suspicious of a method caching you could uncomment this out:
        # attempts.extend([tuple(G) for G in itertools.permutations(flat, 3)])
        skip = set()

        # collect functions which mutate arrays but don't change our hash
        broken = []

        with warnings.catch_warnings():
            # ignore all warnings inside this context manager
            warnings.filterwarnings("ignore")

            for method in list(dir(tracked_array(g.random(dim)))):
                if method in skip:
                    continue

                failures = []
                g.log.debug(f"hash check: `{method}`")
                for A in attempts:
                    m = g.random((100, 3))
                    true_pre = m.tobytes()
                    m = tracked_array(m)
                    hash_pre = hash(m)
                    try:
                        eval(f"m.{method}(*A)")
                    except BaseException as J:
                        failures.append(str(J))

                    hash_post = hash(m)
                    true_post = m.tobytes()

                    # if tobytes disagrees with our hashing logic
                    # it indicates we have cached incorrectly
                    if (hash_pre == hash_post) != (true_pre == true_post):
                        broken.append((method, A))

        if len(broken) > 0:
            method_busted = {method for method, _ in broken}
            raise ValueError(
                f"`TrackedArray` incorrectly hashing methods: {method_busted}"
            )

    def test_validate(self):
        # create a mesh with two duplicate triangles
        # and one degenerate triangle
        m = g.trimesh.Trimesh(
            vertices=[[0, 0, 0], [1, 0, 0], [0, 1, 0], [1, 0, 0], [0, 1, 0], [1, 1, 0]],
            faces=[[3, 4, 4], [0, 1, 2], [0, 1, 2]],
            validate=False,
        )

        # should not have removed any triangles
        assert m.triangles.shape == (3, 3, 3)
        # should remove every triangle except one
        m.process(validate=True)
        assert m.triangles.shape == (1, 3, 3)

    def test_smooth_shade(self, count=10):
        # test to make sure the smooth shaded copy is cached correctly
        mesh = g.trimesh.creation.cylinder(radius=1, height=10)
        scene = g.trimesh.Scene({"mesh": mesh})

        initial = scene.camera_transform.copy()

        hashes = set()
        for n in range(count):
            angle = g.np.pi * n / count
            matrix = g.trimesh.transformations.rotation_matrix(angle, [1, 0, 0])
            scene.geometry["mesh"].apply_transform(matrix)
            hashes.add(hash(scene.geometry["mesh"].smooth_shaded))
            scene.camera_transform = initial

        # the smooth shade should be unique for every transform
        assert len(hashes) == count


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    g.unittest.main()



try:
    from . import generic as g
except BaseException:
    import generic as g


class IdentifierTest(g.unittest.TestCase):
    def test_identifier(self, count=25):
        meshes = g.np.append(
            list(g.get_meshes(only_watertight=True, split=True, min_volume=0.001)),
            g.get_mesh("fixed_top.ply"),
        )
        for mesh in meshes:
            if not mesh.is_volume or mesh.body_count != 1:
                g.log.warning("Mesh %s is not watertight!", mesh.metadata["file_name"])
                continue

            g.log.info("Trying hash at %d random transforms", count)
            hashed = []
            identifier = []
            for transform in g.random_transforms(count):
                permutated = mesh.copy().apply_transform(transform)
                hashed.append(permutated.identifier_hash)
                identifier.append(permutated.identifier)

            result = g.np.array(hashed)
            ok = (result[0] == result[1:]).all()

            if not ok:
                ptp = g.np.ptp(identifier, axis=0)
                g.log.error(
                    "Hashes on %s differ after transform:\n %s\n",
                    mesh.metadata["file_name"],
                    str(ptp),
                )
                raise ValueError("values differ after transform!")

            # stretch the mesh by a small amount
            stretched = mesh.copy().apply_scale([0.99974507, 0.9995662, 1.0029832])
            if hashed[-1] == stretched.identifier_hash:
                raise ValueError(
                    "Hashes on %s didn't change after stretching",
                    mesh.metadata["file_name"],
                )

    def test_scene_id(self):
        """
        A scene has a nicely constructed transform tree, so
        make sure transforming meshes around it doesn't change
        the nuts of their identifier hash.
        """
        scenes = [g.get_mesh("cycloidal.3DXML")]

        for s in scenes:
            for geom_name, mesh in s.geometry.items():
                meshes = []
                if not mesh.is_volume:
                    continue

                for node in s.graph.nodes_geometry:
                    T, geo = s.graph[node]
                    if geom_name != geo:
                        continue

                    m = mesh.copy()
                    m.apply_transform(T)
                    meshes.append(m)
                if not all(
                    meshes[0].identifier_hash == i.identifier_hash for i in meshes
                ):
                    raise ValueError(f"{geom_name} differs after transform!")

        # check an example for a mirrored part
        assert (
            scenes[0].geometry["disc_cam_B"].identifier_hash
            != scenes[0].geometry["disc_cam_A"].identifier_hash
        )

    def test_reflection(self):
        # identifier should detect mirroring
        a = g.get_mesh("featuretype.STL")
        b = a.copy()
        b.vertices[:, 2] *= -1.0
        b.invert()
        assert g.np.isclose(a.volume, b.volume)
        # hash should differ
        assert a.identifier_hash != b.identifier_hash

        # a mesh which is quite sensitive to mirroring
        a = g.get_mesh("mirror.ply")
        b = a.copy()
        b.vertices[:, 2] *= -1.0
        b.invert()
        assert g.np.isclose(a.volume, b.volume)
        # hash should differ
        assert a.identifier_hash != b.identifier_hash

    def test_duplicates(self):
        def clean_name(name):
            return name.split("_", 1)[0].split("#", 1)[0]

        # a scene with instances
        s = g.get_mesh("cycloidal.3DXML")
        # a flat scene dump
        d = g.trimesh.Scene(s.dump())

        # should have the same number of geometry nodes
        assert len(s.graph.nodes_geometry) == len(d.graph.nodes_geometry)

        # uses the identifier to calculate
        a = s.duplicate_nodes
        b = d.duplicate_nodes

        # should be the same in both forms
        assert len(a) == len(b)

        a_set = {tuple(sorted([clean_name(i) for i in group])) for group in a}
        b_set = {tuple(sorted([clean_name(i) for i in group])) for group in b}
        assert a_set == b_set

        ptp = []
        for group in b:
            current = []
            for node in group:
                _, geom = d.graph[node]
                current.append(d.geometry[geom].identifier)
            ptp.append(g.np.ptp(current, axis=0))
            assert (ptp[-1] < 0.01).all()


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    g.unittest.main()



try:
    from . import generic as g
except BaseException:
    import generic as g

try:
    import gmsh
except ImportError:
    gmsh = None


class GMSHTest(g.unittest.TestCase):
    def test_generate(self):
        # don't test gmsh if not installed
        if gmsh is None:
            return
        # create volume mesh of simple icosahedron
        m = g.trimesh.creation.icosahedron()
        result = g.trimesh.interfaces.gmsh.to_volume(m)
        assert len(result) > 0


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    g.unittest.main()



"""
boolean.py
-------------

Do boolean operations on meshes using either Blender or Manifold.
"""
import warnings

import numpy as np

try:
    from manifold3d import Manifold, Mesh
except BaseException as E:
    from .exceptions import ExceptionWrapper

    Mesh = ExceptionWrapper(E)
    Manifold = ExceptionWrapper(E)

from . import interfaces


def difference(meshes, engine=None, **kwargs):
    """
    Compute the boolean difference between a mesh an n other meshes.

    Parameters
    ----------
    meshes : list of trimesh.Trimesh
      Meshes to be processed
    engine : str
      Which backend to use, i.e. 'blender' or 'manifold'

    Returns
    ----------
    difference : a - (other meshes), **kwargs for a Trimesh
    """
    result = _engines[engine](meshes, operation="difference", **kwargs)
    return result


def union(meshes, engine=None, **kwargs):
    """
    Compute the boolean union between a mesh an n other meshes.

    Parameters
    ----------
    meshes : list of trimesh.Trimesh
      Meshes to be processed
    engine : str
      Which backend to use, i.e. 'blender' or 'manifold'

    Returns
    ----------
    union : a + (other meshes), **kwargs for a Trimesh
    """
    result = _engines[engine](meshes, operation="union", **kwargs)
    return result


def intersection(meshes, engine=None, **kwargs):
    """
    Compute the boolean intersection between a mesh an n other meshes.

    Parameters
    ----------
    meshes : list of trimesh.Trimesh
      Meshes to be processed
    engine : str
      Which backend to use, i.e. 'blender' or 'manifold'

    Returns
    ----------
    intersection : **kwargs for a Trimesh object of the
                    volume that is contained by all meshes
    """
    result = _engines[engine](meshes, operation="intersection", **kwargs)
    return result


def boolean_manifold(meshes, operation, debug=False, **kwargs):
    """
    Run an operation on a set of meshes using the Manifold engine.
    """
    # Convert to manifold meshes
    manifolds = [
        Manifold.from_mesh(
            Mesh(
                vert_properties=np.asarray(mesh.vertices, dtype="float32"),
                tri_verts=np.asarray(mesh.faces, dtype="int32"),
            )
        )
        for mesh in meshes
    ]

    # Perform operations
    if operation == "difference":
        if len(meshes) != 2:
            raise ValueError("Difference only defined over two meshes.")

        result_manifold = manifolds[0] - manifolds[1]
    elif operation == "union":
        result_manifold = manifolds[0]

        for manifold in manifolds[1:]:
            result_manifold = result_manifold + manifold
    elif operation == "intersection":
        result_manifold = manifolds[0]

        for manifold in manifolds[1:]:
            result_manifold = result_manifold ^ manifold
    else:
        raise ValueError(f"Invalid boolean operation: '{operation}'")

    # Convert back to trimesh meshes
    from . import Trimesh

    result_mesh = result_manifold.to_mesh()
    out_mesh = Trimesh(vertices=result_mesh.vert_properties, faces=result_mesh.tri_verts)

    return out_mesh


def boolean_scad(*args, **kwargs):
    warnings.warn(
        "The OpenSCAD interface is deprecated, and Trimesh will instead"
        " use Manifold ('manifold'), which should be equivalent. In future versions"
        " of Trimesh, attempting to use engine 'scad' may raise an error.",
        DeprecationWarning,
        stacklevel=2,
    )
    return boolean_manifold(*args, **kwargs)


# which backend boolean engines
_engines = {
    None: boolean_manifold,
    "auto": boolean_manifold,
    "manifold": boolean_manifold,
    "scad": boolean_scad,
    "blender": interfaces.blender.boolean,
}



try:
    from . import generic as g
except BaseException:
    import generic as g


class ThicknessTest(g.unittest.TestCase):
    def test_sphere_thickness(self):
        m = g.trimesh.creation.box()

        samples, faces = m.sample(1000, return_index=True)

        # check that thickness measures are non-negative
        thickness = g.trimesh.proximity.thickness(
            mesh=m,
            points=samples,
            exterior=False,
            normals=m.face_normals[faces],
            method="max_sphere",
        )
        assert (thickness > -g.trimesh.tol.merge).all()

        # check thickness at a specific point
        point = g.np.array([[0.5, 0.0, 0.0]])
        point_thickness = g.trimesh.proximity.thickness(
            m, point, exterior=False, method="max_sphere"
        )

        # its a unit cube
        assert g.np.allclose(point_thickness, 1.0)

    def test_ray_thickness(self):
        m = g.trimesh.creation.box()

        samples, faces = m.sample(1000, return_index=True)

        # check that thickness and measures are non-negative
        thickness = g.trimesh.proximity.thickness(
            mesh=m,
            points=samples,
            exterior=False,
            normals=m.face_normals[faces],
            method="ray",
        )

        assert (thickness > -g.trimesh.tol.merge).all()

        # check thickness at a specific point
        point = g.np.array([[0.5, 0.0, 0.0]])
        point_thickness = g.trimesh.proximity.thickness(
            mesh=m, points=point, exterior=False, method="ray"
        )
        assert g.np.allclose(point_thickness, 1.0)

    def test_sphere_reach(self):
        m = g.trimesh.creation.box()

        samples, faces = m.sample(1000, return_index=True)

        # check that reach measures are infinite
        reach = g.trimesh.proximity.thickness(
            mesh=m,
            points=samples,
            exterior=True,
            normals=m.face_normals[faces],
            method="max_sphere",
        )
        assert g.np.isinf(reach).all()

    def test_ray_reach(self):
        m = g.trimesh.creation.box()

        samples, faces = m.sample(1000, return_index=True)

        # check that reach measures are infinite
        reach = g.trimesh.proximity.thickness(
            mesh=m,
            points=samples,
            exterior=True,
            normals=m.face_normals[faces],
            method="ray",
        )
        assert g.np.isinf(reach).all()

    def test_known(self):
        # an axis aligned plate part
        m = g.get_mesh("1002_tray_bottom.STL")

        # points on the surface
        samples = m.sample(1000)

        # compute thicknesses using sphere method
        rs = g.trimesh.proximity.thickness(mesh=m, points=samples, method="max_sphere")
        # compute thicknesses using pure ray tests
        ra = g.trimesh.proximity.thickness(mesh=m, points=samples, method="ray")
        # mesh is axis aligned plate so thickness is min extent
        truth = m.extents.min()

        assert g.np.isclose(g.np.median(rs), truth)
        assert g.np.isclose(g.np.median(ra), truth)


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    g.unittest.main()



try:
    from . import generic as g
except BaseException:
    import generic as g


class TrianglesTest(g.unittest.TestCase):
    def test_barycentric(self):
        for m in g.get_meshes(4):
            # a simple test which gets the barycentric coordinate at each of the three
            # vertices, checks to make sure the barycentric is [1,0,0] for the vertex
            # and then converts back to cartesian and makes sure the original points
            #  are the same as the conversion and back
            for method in ["cross", "cramer"]:
                for i in range(3):
                    barycentric = g.trimesh.triangles.points_to_barycentric(
                        m.triangles, m.triangles[:, i], method=method
                    )
                    assert (
                        g.np.abs(barycentric - g.np.roll([1.0, 0, 0], i)) < 1e-8
                    ).all()

                    points = g.trimesh.triangles.barycentric_to_points(
                        m.triangles, barycentric
                    )
                    assert (g.np.abs(points - m.triangles[:, i]) < 1e-8).all()

    def test_closest(self):
        closest = g.trimesh.triangles.closest_point(
            triangles=g.data["triangles"]["triangles"],
            points=g.data["triangles"]["points"],
        )

        comparison = (closest - g.data["triangles"]["closest"]).all()

        assert (comparison < 1e-8).all()
        g.log.info("finished closest check on %d triangles", len(closest))

    def test_closest_obtuse(self):
        # simple triangle in the xy-plane with an obtuse corner at vertex A
        ABC = g.np.float32([[0, 0, 0], [2, 0, 0], [-2, 1, 0]])
        D = g.np.float32([1, -1, 0])

        # ground truth: closest point from D is the center of the AB edge:
        # (1,0,0)
        gt_closest = g.np.float32([1, 0, 0])
        tm_closest = g.trimesh.triangles.closest_point([ABC], [D])[0]

        assert g.np.linalg.norm(gt_closest - tm_closest) < g.tol.merge

        # create a circle of points around the triangle
        # with a radius so that all points are outside of the triangle
        radius = 3
        nPtsOnCircle = 100
        alphas = g.np.linspace(
            g.np.pi / nPtsOnCircle, g.np.pi * 2 - g.np.pi / nPtsOnCircle, nPtsOnCircle
        )
        ptsOnCircle = (
            g.np.transpose([g.np.cos(alphas), g.np.sin(alphas), g.np.zeros(nPtsOnCircle)])
            * radius
        )

        def norm(v):
            return g.np.sqrt(g.np.einsum("...i,...i", v, v))

        def distToLine(o, v, p):
            return norm((o - p) - g.np.dot(o - p, v) * v)

        def distPointToEdge(U, V, P):  # edge [U, V], point P
            UtoV = V - U
            UtoP = P - U
            VtoP = P - V

            if g.np.dot(UtoV, UtoP) <= 0:
                # P is 'behind' U
                return norm(UtoP)
            elif g.np.dot(-UtoV, VtoP) <= 0:
                # P is 'behind' V
                return norm(VtoP)
            else:
                # P is 'between' U and V
                return distToLine(U, UtoV / norm(UtoV), P)

        # get closest points from trimesh and compute distances to the circle
        # points
        tm_dists = norm(
            ptsOnCircle
            - g.trimesh.triangles.closest_point([ABC] * nPtsOnCircle, ptsOnCircle)
        )
        # compute naive point-to-edge distances for all points and take the min of
        # the three edges
        gt_dists = g.np.float32(
            [
                [distPointToEdge(ABC[i], ABC[(i + 1) % 3], pt) for i in range(3)]
                for pt in ptsOnCircle
            ]
        ).min(axis=1)
        diff_dists = tm_dists - gt_dists

        assert g.np.dot(diff_dists, diff_dists) < g.tol.merge

    def test_degenerate(self):
        tri = [
            [[0, 0, 0], [1, 0, 0], [-0.5, 0, 0]],
            [[0, 0, 0], [0, 0, 0], [10, 10, 0]],
            [[0, 0, 0], [0, 0, 2], [0, 0, 2.2]],
            [[0, 0, 0], [1, 0, 0], [0, 1, 0]],
        ]

        tri_gt = [False, False, False, True]

        r = g.trimesh.triangles.nondegenerate(tri)
        assert len(r) == len(tri)
        assert (r == tri_gt).all()

    def test_angles(self):
        # a zero- area triangle
        tris = g.np.array([[[0, 0, 0], [1, 0, 0], [1, 0, 0]]], dtype=g.np.float64)
        angles = g.trimesh.triangles.angles(tris)
        # degenerate angles should be zero, not NaN
        g.log.debug(angles)
        assert g.np.allclose(angles, 0.0)

        # an equilateral triangle
        tris = g.np.array(
            [[[-1, 0, 0], [1, 0, 0], [0, g.np.sqrt(3), 0]]], dtype=g.np.float64
        )
        angles = g.trimesh.triangles.angles(tris)
        # degenerate angles should be zero, not NaN
        assert g.np.allclose(angles, g.np.radians(60))

        # an equilateral triangle transformed into space
        tris = g.trimesh.transform_points(
            g.np.array([[-1, 0, 0], [1, 0, 0], [0, g.np.sqrt(3), 0]], dtype=g.np.float64),
            g.trimesh.transformations.random_rotation_matrix(),
        ).reshape((-1, 3, 3))
        angles = g.trimesh.triangles.angles(tris)
        # all angles should be 60 degrees
        assert g.np.allclose(angles, g.np.radians(60))

        # an 3-4-5 right triangle
        tris = g.trimesh.transform_points(
            g.np.array([[0, 0, 0], [3, 0, 0], [0, 4, 0]], dtype=g.np.float64),
            g.trimesh.transformations.random_rotation_matrix(),
        ).reshape((-1, 3, 3))
        # get angles
        angles = g.trimesh.triangles.angles(tris)
        # make sure they match a 3-4-5
        assert g.np.allclose(
            g.np.sort(angles.ravel()),
            [g.np.arcsin(3.0 / 5), g.np.arcsin(4.0 / 5), g.np.pi / 2],
        )


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    g.unittest.main()



from . import ray_triangle

# optionally load an interface to the embree raytracer
try:
    from . import ray_pyembree

    has_embree = True
except BaseException as E:
    from .. import exceptions

    ray_pyembree = exceptions.ExceptionWrapper(E)
    has_embree = False

# add to __all__ as per pep8
__all__ = ["ray_triangle", "ray_pyembree"]



try:
    from . import generic as g
except BaseException:
    import generic as g


class RenderTest(g.unittest.TestCase):
    """
    Test to make sure we can convert trimesh objects
    to pyglet objects for rendering.
    """

    def test_args(self):
        # not imported to trimesh by __init__
        # so we don't have a pyglet import attempt
        # unless we explicitly ask for it (like during
        # a viewer show() call)
        from trimesh import rendering

        files = ["featuretype.STL", "fuze.obj", "points_bin.ply"]
        for file_name in files:
            m = g.get_mesh(file_name)
            args = rendering.convert_to_vertexlist(m)

            if isinstance(m, g.trimesh.Trimesh):
                # try turning smoothing off and on
                rendering.mesh_to_vertexlist(m, smooth_threshold=0)
                rendering.mesh_to_vertexlist(m, smooth_threshold=g.np.inf)

                P30 = m.section(plane_normal=[0, 0, 1], plane_origin=m.centroid)
                args = rendering.path_to_vertexlist(P30)
                args_auto = rendering.convert_to_vertexlist(P30)
                assert len(args) == 6
                assert len(args_auto) == len(args)

                P20, T = P30.to_planar()
                args = rendering.path_to_vertexlist(P20)
                args_auto = rendering.convert_to_vertexlist(P20)
                assert len(args) == 6
                assert len(args_auto) == len(args)

        P21 = g.get_mesh("2D/wrench.dxf")
        args = rendering.path_to_vertexlist(P21)
        args_auto = rendering.convert_to_vertexlist(P21)
        assert len(args) == 6
        assert len(args_auto) == len(args)

        P22 = g.random((100, 2))
        args = rendering.points_to_vertexlist(P22)
        args_auto = rendering.convert_to_vertexlist(P22)
        assert len(args) == 6
        assert len(args_auto) == len(args)

        P31 = g.random((100, 3))
        args = rendering.points_to_vertexlist(P31)
        args_auto = rendering.convert_to_vertexlist(P31)
        assert len(args) == 6
        assert len(args_auto) == len(args)

        P32 = g.trimesh.points.PointCloud(P31)
        args = rendering.convert_to_vertexlist(P32)
        assert len(args) == 6
        assert len(args_auto) == len(args)


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    g.unittest.main()



try:
    from . import generic as g
except BaseException:
    import generic as g


class OBJTest(g.unittest.TestCase):
    def test_rabbit(self):
        # A BSD-licensed test model from pyglet
        # it has mixed triangles, quads, and 16 element faces -_-
        # this should test the non-vectorized load path
        m = g.get_mesh("rabbit.obj")
        assert len(m.faces) == 1252
        rec = g.wrapload(m.export(file_type="obj"), file_type="obj")
        assert g.np.isclose(m.area, rec.area)

    def test_no_img(self):
        # sometimes people use the `vt` parameter for arbitrary
        # vertex attributes and thus want UV coordinates even
        # if there is no texture image
        m = g.get_mesh("noimg.obj")
        assert m.visual.uv.shape == (len(m.vertices), 2)
        # make sure UV coordinates are in range 0.0 - 1.0
        assert m.visual.uv.max() < (1 + 1e-5)
        assert m.visual.uv.min() > -1e-5
        # check to make sure it's not all zeros
        assert m.visual.uv.ptp() > 0.5
        rec = g.wrapload(m.export(file_type="obj"), file_type="obj")
        assert g.np.isclose(m.area, rec.area)

    def test_trailing(self):
        # test files with texture and trailing slashes
        m = g.get_mesh("jacked.obj")
        assert len(m.visual.uv) == len(m.vertices)
        rec = g.wrapload(m.export(file_type="obj"), file_type="obj")
        assert g.np.isclose(m.area, rec.area)

    def test_obj_groups(self):
        # a wavefront file with groups defined
        mesh = g.get_mesh("groups.obj")

        # make sure some data got loaded
        assert g.trimesh.util.is_shape(mesh.faces, (-1, 3))
        assert g.trimesh.util.is_shape(mesh.vertices, (-1, 3))

        # make sure groups are the right length
        # TODO: we do not support face groups now
        # assert len(mesh.metadata['face_groups']) == len(mesh.faces)

        # check to make sure there is signal not just zeros
        # assert mesh.metadata['face_groups'].ptp() > 0

    def test_obj_negative_indices(self):
        # a wavefront file with negative indices
        mesh = g.get_mesh("negative_indices.obj")

        # make sure some data got loaded
        assert g.trimesh.util.is_shape(mesh.faces, (12, 3))
        assert g.trimesh.util.is_shape(mesh.vertices, (8, 3))

    def test_obj_quad(self):
        mesh = g.get_mesh("quadknot.obj")
        # make sure some data got loaded
        assert g.trimesh.util.is_shape(mesh.faces, (-1, 3))
        assert g.trimesh.util.is_shape(mesh.vertices, (-1, 3))

        assert mesh.is_watertight
        assert mesh.is_winding_consistent
        rec = g.wrapload(mesh.export(file_type="obj"), file_type="obj")
        assert g.np.isclose(mesh.area, rec.area)

    def test_obj_multiobj(self):
        # test a wavefront file with multiple objects in the same file
        scene = g.get_mesh("two_objects.obj", split_object=True, group_material=False)
        assert len(scene.geometry) == 2

        for mesh in scene.geometry.values():
            # make sure some data got loaded
            assert g.trimesh.util.is_shape(mesh.faces, (-1, 3))
            assert g.trimesh.util.is_shape(mesh.vertices, (-1, 3))

            assert mesh.is_watertight
            assert mesh.is_winding_consistent

    def test_obj_split_attributes(self):
        # test a wavefront file where pos/uv/norm have different indices
        # and where multiple objects share vertices
        # Note 'process=False' to avoid merging vertices
        scene = g.get_mesh(
            "joined_tetrahedra.obj",
            process=False,
            split_object=True,
            group_material=False,
        )

        assert len(scene.geometry) == 2

        geom = list(scene.geometry.values())

        assert g.trimesh.util.is_shape(geom[0].faces, (4, 3))
        assert g.trimesh.util.is_shape(geom[0].vertices, (9, 3))
        assert g.trimesh.util.is_shape(geom[1].faces, (4, 3))
        assert g.trimesh.util.is_shape(geom[1].vertices, (9, 3))

    def test_obj_simple_order(self):
        # test a simple wavefront model without split indexes
        # and make sure we don't reorder vertices unnecessarily
        file_name = g.os.path.join(g.dir_models, "cube.OBJ")
        # load a simple OBJ file without merging vertices
        m = g.trimesh.load(file_name, process=False)
        # use trivial loading to compare with fancy performant one
        with open(file_name) as f:
            f, v, vt = simple_load(f.read())
        # trimesh loader should return the same face order
        assert g.np.allclose(f, m.faces)
        assert g.np.allclose(v, m.vertices)

    def test_order_tex(self):
        # test a simple wavefront model without split indexes
        # and make sure we don't reorder vertices unnecessarily
        file_name = g.os.path.join(g.dir_models, "fuze.obj")
        # load a simple OBJ file without merging vertices
        m = g.trimesh.load(file_name, process=False, maintain_order=True)
        # use trivial loading to compare with fancy performant one
        with open(file_name) as f:
            f, v, vt = simple_load(f.read())
        # trimesh loader should return the same face order
        assert g.np.allclose(f, m.faces)
        assert g.np.allclose(v, m.vertices)

    def test_obj_compressed(self):
        mesh = g.get_mesh("cube_compressed.obj", process=False)
        assert mesh._cache.cache["vertex_normals"].shape == mesh.vertices.shape
        assert g.np.allclose(g.np.abs(mesh.vertex_normals).sum(axis=1), 1.0)

    def test_vertex_color(self):
        # get a box mesh
        mesh = g.trimesh.creation.box()
        # set each vertex to a unique random color
        mesh.visual.vertex_colors = [
            g.trimesh.visual.random_color() for _ in range(len(mesh.vertices))
        ]
        # export and then reload the file as OBJ
        rec = g.trimesh.load(
            g.trimesh.util.wrap_as_stream(mesh.export(file_type="obj")), file_type="obj"
        )
        # assert colors have survived the export cycle
        assert (mesh.visual.vertex_colors == rec.visual.vertex_colors).all()

    def test_single_vn(self):
        """
        Make sure files with a single VN load.
        """
        m = g.get_mesh("singlevn.obj")
        assert len(m.vertices) > 0
        assert len(m.faces) > 0

    def test_polygon_faces(self):
        m = g.get_mesh("polygonfaces.obj")
        assert len(m.vertices) > 0
        assert len(m.faces) > 0

    def test_faces_not_enough_indices(self):
        m = g.get_mesh("notenoughindices.obj")
        assert len(m.vertices) > 0
        assert len(m.faces) == 1

    def test_export_path(self):
        m = g.get_mesh("fuze.obj")
        g.check_fuze(m)
        assert m._cache.cache["vertex_normals"].shape == m.vertices.shape
        with g.TemporaryDirectory() as d:
            file_path = g.os.path.join(d, "fz.obj")
            m.export(file_path)
            r = g.trimesh.load(file_path)
            g.check_fuze(r)

    def test_skip_mtl(self):
        # not loading materials should produce a trivial texture
        m_tex = g.get_mesh("fuze.obj")
        m_tex_size = m_tex.visual.material.image.size

        m_notex = g.get_mesh("fuze.obj", skip_materials=True)
        m_notex_size = m_notex.visual.material.image.size

        assert m_tex_size != m_notex_size

    def test_mtl(self):
        # get a mesh with texture
        m = g.get_mesh("fuze.obj")
        # export the mesh including data
        obj, data = g.trimesh.exchange.export.export_obj(m, return_texture=True)
        with g.TemporaryDirectory() as path:
            # where is the OBJ file going to be saved
            obj_path = g.os.path.join(path, "test.obj")
            with open(obj_path, "w") as f:
                f.write(obj)
            # save the MTL and images
            for k, v in data.items():
                with open(g.os.path.join(path, k), "wb") as f:
                    f.write(v)
            # reload the mesh from the export
            rec = g.trimesh.load(obj_path)
        # make sure loaded image is the same size as the original
        assert rec.visual.material.image.size == m.visual.material.image.size
        # make sure the faces are the same size
        assert rec.faces.shape == m.faces.shape

    def test_scene(self):
        s = g.get_mesh("cycloidal.3DXML")

        e = g.trimesh.load(
            g.io_wrap(s.export(file_type="obj")),
            file_type="obj",
            split_object=True,
            group_materials=False,
        )

        assert g.np.isclose(e.area, s.area, rtol=0.01)

    def test_edge_cases(self):
        # a mesh with some NaN colors
        n = g.get_mesh("nancolor.obj")
        assert n.faces.shape == (12, 3)

        v = g.get_mesh("cubevt.obj")
        assert v.faces.shape == (12, 3)

    def test_empty_or_pointcloud(self):
        # demo files to check
        empty_files = ["obj_empty.obj", "obj_points.obj", "obj_wireframe.obj"]

        for empty_file in empty_files:
            e = g.get_mesh("emptyIO/" + empty_file)

            # create export
            if "empty" in empty_file:
                try:
                    export = e.export(file_type="ply")
                except BaseException:
                    continue
                raise ValueError("cannot export empty")
            elif "points" in empty_file:
                export = e.export(file_type="ply")
                reconstructed = g.wrapload(export, file_type="ply")

                # result should be a point cloud instance
                assert isinstance(e, g.trimesh.PointCloud)
                assert hasattr(e, "vertices")
                # point cloud export should contain vertices
                assert isinstance(reconstructed, g.trimesh.PointCloud)
                assert hasattr(reconstructed, "vertices")

    def test_backslash_continuation_character(self):
        # an obj file with \ (backslash) line continuation characters
        m = g.get_mesh("wallhole.obj")
        assert m.faces.shape == (66, 3)

    def test_no_uv(self):
        mesh = g.get_mesh("box.obj")
        rec = g.wrapload(mesh.export(file_type="obj"), file_type="obj")
        assert g.np.isclose(mesh.area, rec.area)

    def test_no_uv_but_mtl(self):
        sphere = g.trimesh.creation.uv_sphere()
        sphere.visual = g.trimesh.visual.TextureVisuals(
            uv=None, material=g.trimesh.visual.material.empty_material()
        )
        output = sphere.export("sphere.obj")
        assert "usemtl" in output

    def test_chair(self):
        mesh = next(iter(g.get_mesh("chair.zip").geometry.values()))

        # this model comes with vertex normals
        assert "vertex_normals" in mesh._cache
        assert g.np.allclose(1.0, g.np.linalg.norm(mesh.vertex_normals, axis=1))
        mesh.apply_scale(0.46377314288075433)
        assert "vertex_normals" in mesh._cache
        assert g.np.allclose(1.0, g.np.linalg.norm(mesh.vertex_normals, axis=1))
        assert "vertex_normals" in mesh._cache
        mesh._cache.clear()
        assert "vertex_normals" not in mesh._cache
        # if we recomputed now, the degenerate faces
        # would lead some of these vertex normals to be zero
        # assert g.np.allclose(
        #    1.0, g.np.linalg.norm(mesh.vertex_normals, axis=1))

    def test_multi_nodupe(self):
        s = g.get_mesh("forearm.zae")
        obj, mtl = g.trimesh.exchange.obj.export_obj(
            s, include_color=True, include_texture=True, return_texture=True
        )
        # should be using one material file
        assert obj.count("mtllib") == 1
        assert "mtllib material.mtl" in obj
        # should be specifying 5 materials
        assert obj.count("usemtl") == 5

        # this file has only the properties (no images)
        assert len(mtl) == 1
        mtl_names = [
            L.strip().split()[-1].strip()
            for L in mtl["material.mtl"].decode("utf-8").split("\n")
            if "newmtl" in L
        ]
        # there should be 5 unique material names
        assert len(set(mtl_names)) == 5

    def test_mtl_color_roundtrip(self):
        # create a mesh with a simple material
        m = g.trimesh.creation.box()
        m.visual = m.visual.to_texture()
        # set each color component to a unique value
        colors = [g.trimesh.visual.color.random_color() for _ in range(3)]
        m.visual.material.ambient = colors[0]
        m.visual.material.specular = colors[1]
        m.visual.material.diffuse = colors[2]
        m.visual.material.glossiness = 0.52622

        with g.TemporaryDirectory() as d:
            # exporting by filename will automatically
            # create a FilePathResolver which writes the
            # `mtl` file to the same directory
            file_name = g.os.path.join(d, "hi.obj")
            m.export(file_name)
            # reload the export by file name
            r = g.trimesh.load(file_name)

        # these values should have survived the roundtrip
        assert g.np.allclose(m.visual.material.ambient, r.visual.material.ambient)
        assert g.np.allclose(m.visual.material.specular, r.visual.material.specular)
        assert g.np.allclose(m.visual.material.diffuse, r.visual.material.diffuse)
        assert g.np.isclose(m.visual.material.glossiness, r.visual.material.glossiness)

    def test_scene_export_material_name(self):
        s = g.get_mesh("fuze.obj", force="scene")
        dummy = "fuxx"
        s.geometry["fuze.obj"].visual.material.name = dummy

        r = g.trimesh.resolvers.ZipResolver()
        r["model.obj"] = s.export(file_type="obj", mtl_name="mystuff.mtl", resolver=r)

        mtl = r["mystuff.mtl"].decode("utf-8")
        assert mtl.count("newmtl") == 1
        assert f"newmtl {dummy}" in mtl
        assert f"{dummy}.jpeg" in r

    def test_compound_scene_export(self):
        # generate a mesh with multiple textures
        a = g.get_mesh("BoxTextured.glb")
        a = a.scaled(1.0 / a.extents.max())
        a.apply_translation(-a.bounds[0])

        b = g.get_mesh("fuze.obj").scene()
        b = b.scaled(1.0 / b.extents.max())
        b.apply_translation(-b.bounds[0] + [2, 0, 0])

        d = next(iter(b.copy().geometry.values()))
        d.apply_translation([-1, 0, 0])
        assert hash(d.visual.material) == hash(b.geometry["fuze.obj"].visual.material)

        # should change the material hash
        d.visual.material.glossiness = 0.1
        assert hash(d.visual.material) != hash(b.geometry["fuze.obj"].visual.material)

        # generate a compound scene
        c = a + b + d
        for i in c.geometry.values():
            # name all the materials the same thing
            i.visual.material.name = "material_0"

        # export the compound scene
        obj, mtl = c.export(file_type="obj", return_texture=True)
        # there should be exactly one mtllib referenced
        assert obj.count("mtllib") == 1
        assert obj.count("usemtl") == 3

        # should be one texture image for each of 3
        # plus the `.mtl` file itself
        # if we had image-hash-deduplication this should
        # be changed to 3 as the image for `b` and `d` are the same
        assert len(mtl) == 4

        # get the material names specified

        mtl_names = [
            L.strip().split()[-1].strip()
            for L in mtl["material.mtl"].decode("utf-8").split("\n")
            if "newmtl" in L
        ]
        # there should be 3 unique material names
        assert len(set(mtl_names)) == 3

        # now reload the compound scene
        t = g.trimesh.load(
            file_obj=g.trimesh.util.wrap_as_stream(obj),
            file_type="obj",
            resolver=g.trimesh.resolvers.ZipResolver(mtl),
            group_material=False,
            split_object=True,
        )
        # these names should match eventually
        assert len(t.geometry.keys()) == len(c.geometry.keys())
        assert g.np.isclose(t.area, c.area)

    def test_face_parsing_in_group_names(self):
        # Checks that an obj with a g tag containinig a face like name (an 'f '
        # followed by three space separated text chunks, ex: f 1 2 3) does load
        # properly
        m = g.get_mesh("face_in_group_name.obj")
        assert len(m.vertices) == 1

    def test_face_parsing_in_group_names_with_object_tag(self):
        # Checks that an obj with a g tag in the middle of a file,
        # containinig a face like name (an 'f ' followed by three
        # space separated text chunks, ex: f 1 2 3), does load properly
        m = g.get_mesh("face_in_group_name_mid_file.obj")
        assert len(m.vertices) == 5
        assert len(m.faces) == 2

    def test_chunk_parsing_with_no_faces_but_with_f_in_chunk(self):
        # Checks that a chunk with no faces but with 'f ' in it loads properly
        m = g.get_mesh("obj_with_no_face_in_chunk.obj")
        assert len(m.vertices) == 3
        assert len(m.faces) == 1

    def test_export_normals(self):
        m = g.trimesh.creation.box()
        assert "vertex_normals" not in m._cache.cache

        e = m.export(file_type="obj", include_normals=None)
        assert "vn " not in e
        e = m.export(file_type="obj", include_normals=True)
        # should have included normals
        assert "vn " in e
        # should have forced generation of normals
        assert "vertex_normals" in m._cache.cache

        # now that they're in cache include_normals=None should get them
        e = m.export(file_type="obj", include_normals=None)
        assert "vn " in e

        # or skip them if explicitly asked
        e = m.export(file_type="obj", include_normals=False)
        assert "vn " not in e

    def test_export_mtl_args(self):
        mesh = g.trimesh.creation.box()
        # check for a crash with no materials defined
        a, b = g.trimesh.exchange.obj.export_obj(
            mesh, return_texture=True, mtl_name="hi.mtl"
        )


def simple_load(text):
    # we're going to load faces in a basic text way
    # and compare the order from this method to the
    # trimesh loader, to see if we get the same thing
    # note that trimesh's extremely convoluted string
    # wangling is wildly faster than this
    f = []
    v = []
    vt = []
    for line in str.splitlines(text):
        line = line.strip()
        if len(line) < 2:
            continue
        elif line.startswith("f "):
            if "/" in line:
                f.append([int(i.split("/", 1)[0]) for i in line[1:].strip().split()])
            else:
                f.append(line[1:].strip().split())
        elif line.startswith("v "):
            v.append(line[1:].strip().split())
        elif line.startswith("vt "):
            vt.append(line[2:].strip().split())

    # get faces as basic numpy array
    f = g.np.array(f, dtype=g.np.int64) - 1
    v = g.np.array(v, dtype=g.np.float64)
    vt = g.np.array(vt, dtype=g.np.float64)

    return f, v, vt


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    g.unittest.main()



try:
    from . import generic as g
except BaseException:
    import generic as g


class TextureTest(g.unittest.TestCase):
    def test_uv_to_color(self):
        try:
            import PIL.Image
        except ImportError:
            return

        # n_vertices = 100
        uv = g.np.array([[0.25, 0.2], [0.4, 0.5]], dtype=float)
        texture = g.np.arange(96, dtype=g.np.uint8).reshape(8, 4, 3)
        colors = g.trimesh.visual.uv_to_color(uv, PIL.Image.fromarray(texture))

        colors_expected = [[75, 76, 77, 255], [51, 52, 53, 255]]

        assert (colors == colors_expected).all()

    def test_bad_uv(self):
        # get a textured OBJ
        m = g.get_mesh("fuze.obj", force="mesh")
        # add malformed UV coordinates
        m.visual.uv = m.visual.uv[:100]
        m.merge_vertices()

    def test_order_kwarg(self):
        for file_name in ["ico4.obj", "ico4uv.obj"]:
            # get the location of the model file
            file_path = g.get_path(file_name)
            with open(file_path) as f:
                # get the raw ordered vertices from the file with basic string
                # ops
                v_raw = g.np.array(
                    [line[2:].split() for line in f if line.startswith("v ")],
                    dtype=g.np.float64,
                )

            # load them with maximal correspondence captain
            a = g.trimesh.load(file_path, process=False, maintain_order=True)
            # see if we have the same vertices
            assert g.np.allclose(a.vertices, v_raw)

            # load them without process and but without maintaining order
            a = g.trimesh.load(file_path, process=False, maintain_order=False)
            # vertex shape should not be the same
            assert a.vertices.shape != v_raw.shape

    def test_fuze(self):
        # create a local web server to test remote assets
        with g.serve_meshes() as address:
            # see if web resolvers work
            tex = g.trimesh.exchange.load.load_remote(
                url=address + "/fuze.obj", process=False
            )
            g.check_fuze(tex)

            # see if web + zip resolvers work
            scene = g.trimesh.exchange.load.load_remote(
                url=address + "/fuze.zip", process=False
            )

            # zip files get loaded into a scene
            assert len(scene.geometry) == 1
            # scene should just be a fuze bottle
            g.check_fuze(next(iter(scene.geometry.values())))

        # obj with texture, assets should be loaded
        # through a FilePathResolver
        m = g.get_mesh("fuze.obj", process=False)
        g.check_fuze(tex)

        # obj with texture, assets should be loaded
        # through a ZipResolver into a scene
        scene = g.get_mesh("fuze.zip", process=False)

        # zip files get loaded into a scene
        assert len(scene.geometry) == 1
        m = next(iter(scene.geometry.values()))
        g.check_fuze(m)

        # the PLY should have textures defined
        m = g.get_mesh("fuze.ply", process=False)
        g.check_fuze(m)

        # ASCII PLY should have textures defined
        m = g.get_mesh("fuze_ascii.ply", process=False)
        g.check_fuze(m)

        # textured meshes should subdivide OK-ish
        s = m.subdivide()
        assert len(s.visual.uv) == len(s.vertices)

        # load without doing the vertex separation
        # will look like garbage but represents original
        # and skips "disconnect vertices with different UV"
        b = g.get_mesh("fuze.ply", process=False, fix_texture=False)
        assert len(b.vertices) == 502
        assert len(b.visual.uv) == 502

    def test_upsize(self):
        """
        Texture images usually want to have sizes that are powers
        of two so resize textures up to the nearest power of two.
        """
        try:
            from PIL import Image
        except BaseException:
            g.log.warning("no PIL, not testing power_resize!")
            return

        # shortcut for the function
        resize = g.trimesh.visual.texture.power_resize

        img = Image.new("RGB", (10, 20))
        assert img.size == (10, 20)
        assert resize(img).size == (16, 32)
        assert resize(img, square=True).size == (32, 32)

        # check with one value on-size
        img = Image.new("RGB", (10, 32))
        assert img.size == (10, 32)
        assert resize(img).size == (16, 32)
        assert resize(img, square=True).size == (32, 32)

        # check early exit pathOA
        img = Image.new("RGB", (32, 32))
        assert img.size == (32, 32)
        assert resize(img).size == (32, 32)
        assert resize(img, square=True).size == (32, 32)

    def test_concatenate(self):
        # test concatenation with texture
        a = g.get_mesh("fuze.obj")
        b = a.copy()
        b.apply_translation([b.extents[0] * 1.25, 0, 0])

        c = a + b
        assert len(c.vertices) > len(a.vertices)
        assert len(c.visual.uv) == len(c.vertices)
        # should have deduplicated image texture
        assert g.np.allclose(c.visual.material.image.size, a.visual.material.image.size)

    def test_concatentate_multi(self):
        colors = [
            [255, 0, 0, 255],
            [0, 255, 0, 255],
            [0, 0, 255, 255],
            [100, 100, 100, 255],
        ]
        funcs = [
            g.trimesh.creation.box,
            g.trimesh.creation.icosphere,
            g.trimesh.creation.capsule,
        ]

        fuze = g.get_mesh("fuze.obj")
        fuze.apply_scale(1.0 / fuze.extents.max())
        fuze.apply_translation([-2, 0, 0] - fuze.bounds[0])

        meshes = []
        for i, color in enumerate(colors):
            for j, f in enumerate(funcs):
                m = f()
                m.visual.face_colors = color
                # convert color visual to texture
                m.visual = m.visual.to_texture()
                m.apply_translation([i * 2.2, j * 2.2, 0.0])
                meshes.append(m)

        c = g.trimesh.util.concatenate(meshes)
        assert isinstance(c.visual, g.trimesh.visual.TextureVisuals)
        assert len(c.faces) == sum(len(i.faces) for i in meshes)
        assert g.np.prod(c.visual.material.image.size) >= 4

        # convert texture back to color
        roundtrip = c.visual.to_color()
        assert roundtrip.kind == "vertex"
        vertex_c = roundtrip.vertex_colors
        # get the unique colors
        unique = vertex_c[g.trimesh.grouping.unique_rows(vertex_c)[0]]

        # roundtripped colors should be a superset of original colors
        assert {tuple(c) for c in unique}.issuperset({tuple(c) for c in colors})

    def test_to_tex(self):
        m = g.trimesh.creation.box()
        color = [255, 0, 0, 255]
        m.visual.face_colors = color
        # convert color visual to texture
        m.visual = m.visual.to_texture()
        # convert back to color
        m.visual = m.visual.to_color()
        assert g.np.allclose(m.visual.main_color, color)

    def test_uv_none(self):
        # setting UV coordinates to None should work
        m = g.get_mesh("fuze.obj")
        m.visual.uv = None
        assert m.visual.uv is None

        # should still be None on a copy
        c = m.copy()
        assert c.visual.uv is None

    def test_pbr_export(self):
        # try loading a textured box
        m = next(iter(g.get_mesh("BoxTextured.glb").geometry.values()))
        # make sure material copy doesn't crash
        m.visual.copy()

        with g.TemporaryDirectory() as d:
            # exports by path allow files to be written
            path = g.os.path.join(d, "box.obj")
            m.export(path)
            # try reloading
            r = g.trimesh.load(path)
            # make sure material survived
            assert r.visual.material.image.size == (256, 256)

    def test_pbr_material_fusion(self):
        # a scene with 5 cubes each with a different material
        # one having only simple color values
        # one with PBR material values
        # one with PBR textures
        # one with emissive textures
        # and one with specular glossiness color values
        m = g.get_mesh("pbr_cubes_emissive_spec_gloss.zip", force="mesh")

        assert isinstance(m, g.trimesh.Trimesh)

        mat_m = m.visual.material
        assert isinstance(mat_m, g.trimesh.visual.material.PBRMaterial)
        assert mat_m.baseColorTexture is not None
        assert mat_m.metallicRoughnessTexture is not None
        assert mat_m.emissiveTexture is not None


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    g.unittest.main()
